"""
A template to benchmark a model
with :epkg:`asv`. The benchmark can be run through
file :epkg:`run_asv.sh` on Linux or :epkg:`run_asv.bat` on
Windows.

.. warning::
    On Windows, you should avoid cloning the repository
    on a folder with a long full name. Visual Studio tends to
    abide by the rule of the maximum path length even though
    the system is told otherwise.
"""
import numpy  # pylint: disable=W0611
from mlprodict.tools.asv_options_helper import get_opset_number_from_onnx
# Import specific to this model.
from sklearn.cross_decomposition import PLSCanonical  # pylint: disable=C0411

from mlprodict.asv_benchmark import _CommonAsvSklBenchmarkTrainableTransform  # pylint: disable=C0412
from mlprodict.onnx_conv import to_onnx  # pylint: disable=W0611, C0412
from mlprodict.onnxrt import OnnxInference  # pylint: disable=W0611, C0412


class TemplateBenchmarkTrainableTransform(_CommonAsvSklBenchmarkTrainableTransform):
    """
    :epkg:`asv` example for a trainable transform,
    Full template can be found in
    `common_asv_skl.py <https://github.com/sdpython/mlprodict/
    blob/master/mlprodict/asv_benchmark/common_asv_skl.py>`_.
    """
    params = [
        ['skl', 'pyrtc', 'ort'],  # values for runtime
        [1, 10, 100, 1000, 10000],  # values for N
        [4, 20],  # values for nf
        [get_opset_number_from_onnx()],  # values for opset
        ['float', 'double'],  # values for dtype
        [None],  # values for optim
    ]

    # additional parameters

    def setup_cache(self):  # pylint: disable=W0235
        super().setup_cache()

    def _create_model(self):
        return PLSCanonical()



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRunUnary


class Not(OpRunUnary):

    def __init__(self, onnx_node, desc=None, **options):
        OpRunUnary.__init__(self, onnx_node, desc=desc,
                            **options)

    def _run(self, x):  # pylint: disable=W0221
        return (numpy.logical_not(x), )

    def _infer_shapes(self, x):  # pylint: disable=W0221
        return (x.copy(dtype=numpy.bool_), )

    def _infer_types(self, x):  # pylint: disable=W0221
        return (numpy.bool_, )

    def to_python(self, inputs):
        return self._to_python_numpy(inputs, 'logical_not')



"""
@file
@brief Wraps :epkg:`numpy` functions into :epkg:`onnx`.

.. versionadded:: 0.6
"""
import warnings
from .onnx_version import FctVersion
from .onnx_numpy_annotation import get_args_kwargs
from .onnx_numpy_compiler import OnnxNumpyCompiler


class _created_classes:
    """
    Class to store all dynamic classes created by wrappers.
    """

    def __init__(self):
        self.stored = {}

    def append(self, name, cl):
        """
        Adds a class into `globals()` to enable pickling on dynamic
        classes.
        """
        if name in self.stored:
            warnings.warn(  # pragma: no cover
                "Class %r overwritten in\n%r\n---\n%r" % (
                    name, ", ".join(sorted(self.stored)), cl),
                RuntimeWarning)
        self.stored[name] = cl
        globals()[name] = cl


_created_classes_inst = _created_classes()


class wrapper_onnxnumpy:
    """
    Intermediate wrapper to store a pointer
    on the compiler (type: @see cl OnnxNumpyCompiler).

    :param compiled: instance of @see cl OnnxNumpyCompiler

    .. versionadded:: 0.6
    """

    def __init__(self, compiled):
        self.compiled = compiled

    def __call__(self, *args, **kwargs):
        """
        Calls the compiled function with arguments `args`.
        """
        return self.compiled(*args, **kwargs)

    def __getstate__(self):
        """
        Serializes everything but the function which generates
        the ONNX graph, not needed anymore.
        """
        return dict(compiled=self.compiled)

    def __setstate__(self, state):
        """
        Serializes everything but the function which generates
        the ONNX graph, not needed anymore.
        """
        self.compiled = state['compiled']


def onnxnumpy(op_version=None, runtime=None, signature=None):
    """
    Decorator to declare a function implemented using
    :epkg:`numpy` syntax but executed with :epkg:`ONNX`
    operators.

    :param op_version: :epkg:`ONNX` opset version
    :param runtime: `'onnxruntime'` or one implemented by @see cl OnnxInference
    :param signature: it should be used when the function
        is not annoatated.

    Equivalent to `onnxnumpy(arg)(foo)`.

    .. versionadded:: 0.6
    """
    def decorator_fct(fct):
        compiled = OnnxNumpyCompiler(
            fct, op_version=op_version, runtime=runtime,
            signature=signature)
        name = "onnxnumpy_%s_%s_%s" % (fct.__name__, str(op_version), runtime)
        newclass = type(
            name, (wrapper_onnxnumpy,),
            {'__doc__': fct.__doc__, '__name__': name})
        _created_classes_inst.append(name, newclass)
        return newclass(compiled)
    return decorator_fct


def onnxnumpy_default(fct):
    """
    Decorator with options to declare a function implemented
    using :epkg:`numpy` syntax but executed with :epkg:`ONNX`
    operators.

    :param fct: function to wrap

    .. versionadded:: 0.6
    """
    return onnxnumpy()(fct)


class wrapper_onnxnumpy_np:
    """
    Intermediate wrapper to store a pointer
    on the compiler (type: @see cl OnnxNumpyCompiler)
    supporting multiple signatures.

    .. versionadded:: 0.6
    """

    def __init__(self, **kwargs):
        self.fct = kwargs['fct']
        self.signature = kwargs['signature']
        self.fctsig = kwargs.get('fctsig', None)
        self.args, self.kwargs = get_args_kwargs(
            self.fct,
            0 if self.signature is None else self.signature.n_optional)
        self.data = kwargs
        self.signed_compiled = {}

    def __getstate__(self):
        """
        Serializes everything but the function which generates
        the ONNX graph, not needed anymore.
        """
        data_copy = {k: v for k, v in self.data.items() if k != 'fct'}
        return dict(signature=self.signature, args=self.args,
                    kwargs=self.kwargs, data=data_copy,
                    signed_compiled=self.signed_compiled)

    def __setstate__(self, state):
        """
        Restores serialized data.
        """
        for k, v in state.items():
            setattr(self, k, v)

    def __getitem__(self, dtype):
        """
        Returns the instance of @see cl wrapper_onnxnumpy
        mapped to *dtype*.

        :param dtype: numpy dtype corresponding to the input dtype
            of the function
        :return: instance of @see cl wrapper_onnxnumpy
        """
        if not isinstance(dtype, FctVersion):
            raise TypeError(  # pragma: no cover
                "dtype must be of type 'FctVersion' not %s: %s." % (
                    type(dtype), dtype))
        if dtype not in self.signed_compiled:
            self._populate(dtype)
            key = dtype
        else:
            key = dtype
        return self.signed_compiled[key]

    def __call__(self, *args, **kwargs):
        """
        Calls the compiled function assuming the type of the first
        tensor in *args* defines the templated version of the function
        to convert into *ONNX*.
        """
        if len(self.kwargs) == 0:
            others = None
        else:
            others = tuple(kwargs.get(k, self.kwargs[k]) for k in self.kwargs)
        key = FctVersion(  # pragma: no cover
            tuple(a if (a is None or hasattr(a, 'fit'))
                  else a.dtype.type for a in args),
            others)
        return self[key](*args)

    def _populate(self, version):
        """
        Creates the appropriate runtime for function *fct*
        """
        compiled = OnnxNumpyCompiler(
            fct=self.data["fct"], op_version=self.data["op_version"],
            runtime=self.data["runtime"], signature=self.data["signature"],
            version=version, fctsig=self.data.get('fctsig', None))
        name = "onnxnumpy_np_%s_%s_%s_%s" % (
            self.data["fct"].__name__, str(self.data["op_version"]),
            self.data["runtime"], version.as_string())
        newclass = type(
            name, (wrapper_onnxnumpy,),
            {'__doc__': self.data["fct"].__doc__, '__name__': name})

        self.signed_compiled[version] = newclass(compiled)

    def _validate_onnx_data(self, X):
        return X


def onnxnumpy_np(op_version=None, runtime=None, signature=None):
    """
    Decorator to declare a function implemented using
    :epkg:`numpy` syntax but executed with :epkg:`ONNX`
    operators.

    :param op_version: :epkg:`ONNX` opset version
    :param runtime: `'onnxruntime'` or one implemented by @see cl OnnxInference
    :param signature: it should be used when the function
        is not annoatated.

    Equivalent to `onnxnumpy(arg)(foo)`.

    .. versionadded:: 0.6
    """
    def decorator_fct(fct):
        name = "onnxnumpy_nb_%s_%s_%s" % (
            fct.__name__, str(op_version), runtime)
        newclass = type(
            name, (wrapper_onnxnumpy_np,), {
                '__doc__': fct.__doc__,
                '__name__': name,
                '__getstate__': wrapper_onnxnumpy_np.__getstate__,
                '__setstate__': wrapper_onnxnumpy_np.__setstate__})
        _created_classes_inst.append(name, newclass)
        return newclass(
            fct=fct, op_version=op_version, runtime=runtime,
            signature=signature)

    return decorator_fct



"""
@file
@brief Optimisation of :epkg:`ONNX` graphs.
"""
from onnx.helper import make_graph
from ._onnx_optimisation_common import (  # pylint: disable=E0611
    _rename_node_input,
    _rename_node_output,
    _apply_optimisation_on_graph,
    _apply_remove_node_fct_node
)


def onnx_remove_node_identity(onnx_model, recursive=True, debug_info=None, **options):
    """
    Removes as many *Identity* nodes as possible.
    The function looks into every node and subgraphs if
    *recursive* is True for identity node. Unless such a
    node directy connects one input to one output, it will
    be removed and every other node gets its inputs or
    outputs accordingly renamed.

    @param      onnx_model      onnx model
    @param      recursive       looks into subgraphs
    @param      debug_info      debug information (private)
    @param      options         additional options (unused)
    @return                     new onnx _model
    """
    if debug_info is None:
        debug_info = [str(type(onnx_model)).rsplit(
            '.', maxsplit=1)[-1].strip("'>")]
    else:
        debug_info = (debug_info +
                      [str(type(onnx_model)).rsplit('.', maxsplit=1)[-1].strip("'>")])

    if hasattr(onnx_model, 'graph'):
        return _apply_optimisation_on_graph(
            onnx_remove_node_identity, onnx_model,
            recursive=recursive, debug_info=debug_info, **options)

    graph = onnx_model

    inputs = set(i.name for i in graph.input)
    outputs = set(o.name for o in graph.output)

    def retrieve_idnodes(graph, existing_nodes):
        idnodes = []
        for i, exnode in enumerate(existing_nodes):
            if exnode is None:
                continue
            if exnode.op_type == 'Identity':
                input = exnode.input[0]
                output = exnode.output[0]
                idnodes.append((i, exnode, input, output))
        return idnodes

    nodes = list(graph.node)
    rem = 1
    while rem > 0:
        rem = 0
        idnodes = retrieve_idnodes(graph, nodes)
        restart = False
        for i, _, inp, out in idnodes:
            if restart:
                break  # pragma: no cover
            if nodes[i] is None:
                # Already removed.
                continue  # pragma: no cover
            if inp in inputs and out in outputs:
                # Cannot be removed.
                continue
            if not restart and out not in outputs:
                # We cannot change an output name.
                for j in range(len(nodes)):  # pylint: disable=C0200
                    if nodes[j] is None:
                        continue
                    if out in nodes[j].input:
                        nodes[j] = _rename_node_input(nodes[j], out, inp)
                        rem += 1
                        if nodes[j].op_type == 'Identity':
                            restart = True  # pragma: no cover
                nodes[i] = None
                rem += 1
                continue
            if not restart and inp not in inputs and inp not in outputs:
                # We cannot change an input name or an output name.
                for j in range(len(nodes)):  # pylint: disable=C0200
                    if nodes[j] is None:
                        continue
                    if inp in nodes[j].output:
                        nodes[j] = _rename_node_output(nodes[j], inp, out)
                        rem += 1
                        if nodes[j].op_type == 'Identity':
                            restart = True  # pragma: no cover
                    if inp in nodes[j].input:
                        nodes[j] = _rename_node_input(nodes[j], inp, out)
                        rem += 1
                        if nodes[j].op_type == 'Identity':
                            restart = True
                nodes[i] = None
                rem += 1

    if recursive:
        # Handles subgraphs.
        for i in range(len(nodes)):  # pylint: disable=C0200
            node = nodes[i]
            if node is None or not (node.attribute):  # pylint: disable=C0325
                continue
            nodes[i] = _apply_remove_node_fct_node(
                onnx_remove_node_identity,
                node, recursive=True, debug_info=debug_info + [node.name])

    # Finally create the new graph.
    nodes = list(filter(lambda n: n is not None, nodes))
    graph = make_graph(nodes, onnx_model.name,
                       onnx_model.input, onnx_model.output,
                       onnx_model.initializer)

    graph.value_info.extend(onnx_model.value_info)  # pylint: disable=E1101
    return graph



"""
@file
@brief Shortcut to *testing.einsum*.
"""

from .einsum_bench import einsum_benchmark
from .einsum_fct import einsum
from .einsum_impl import decompose_einsum_equation, apply_einsum_sequence
from .einsum_impl_classes import EinsumSubOp, GraphEinsumSubOp
from .einsum_impl_ext import (
    numpy_extended_dot,
    numpy_extended_dot_python,
    numpy_extended_dot_matrix,
    numpy_diagonal)



"""
@file
@brief Rewrites some of the converters implemented in
:epkg:`sklearn-onnx`.
"""
import copy
from onnx.helper import make_tensor
from skl2onnx.common.data_types import guess_numpy_type
from skl2onnx.common._apply_operation import apply_concat, apply_identity
from ...onnx_tools.onnx2py_helper import _var_as_dict, guess_proto_dtype
from ...npy.onnx_version import FctVersion


def new_calculate_sklearn_function_transformer_output_shapes(operator):
    """
    Rewrites the converters implemented in
    :epkg:`sklearn-onnx` to support custom functions
    implemented with :ref:`l-numpy-onnxpy`.
    """
    fct = operator.raw_operator.func
    if hasattr(fct, 'signed_compiled'):
        dtype = guess_numpy_type(operator.inputs[0].type)
        fct = fct[FctVersion((dtype, ), None)]
    if hasattr(fct, 'compiled'):
        compiled = fct.compiled
        if not hasattr(compiled, 'onnx_'):
            raise RuntimeError(  # pragma: no cover
                "Attribute 'onnx_' is missing, function was not "
                "converted to onnx.")
        onx = compiled.onnx_
        graph = onx.graph
        outputs = graph.output

        # Let's assume there is only one output
        # with the same type as the input.
        # Only the shape changes.
        if len(outputs) != 1:
            raise RuntimeError(  # pragma: no cover
                "Only one output is allowed not %d." % len(outputs))
        input_type = operator.inputs[0].type.__class__
        if compiled.meta_.get('signature', None):
            dims = compiled.meta_['signature'].shape_calculator(
                operator.inputs[0].type.shape)
        else:
            N = operator.inputs[0].type.shape[0]
            dims = [N]
            out = outputs[0]
            try:
                extra_dims = out.type.tensor_type.shape.dim
            except AttributeError:  # pragma: no cover
                extra_dims = None
            if extra_dims is not None:
                dims.extend(extra_dims[1:])
        operator.outputs[0].type = input_type(dims)
        return

    if operator.raw_operator.func is not None:
        raise TypeError("FunctionTransformer is not supported unless the "
                        "transform function is of type %r "
                        "wrapped with onnxnumpy." % type(
                            operator.raw_operator.func))
    N = operator.inputs[0].type.shape[0]
    C = 0
    for variable in operator.inputs:
        if variable.type.shape[1] is not None:
            C += variable.type.shape[1]
        else:
            C = None
            break

    operator.outputs[0].type = operator.inputs[0].type.__class__([N, C])


def _copy_attributes(att):
    if hasattr(att, 'value'):
        return att.value
    vt = _var_as_dict(att)
    if vt['type']['kind'] == 'tensor':
        value = vt['value']
        return make_tensor(att.name, guess_proto_dtype(value.dtype),
                           value.shape, value.ravel().tolist())
    if vt['type']['kind'] == 'real':
        return vt['value']
    raise RuntimeError(  # pragma: no cover
        "Unable to copy attribute %r, got %r." % (att, vt))


def new_convert_sklearn_function_transformer(scope, operator, container):
    """
    Rewrites the converters implemented in
    :epkg:`sklearn-onnx` to support custom functions
    implemented with :ref:`l-numpy-onnxpy`.
    """
    op = operator.raw_operator
    fct = op.func
    if hasattr(fct, 'signed_compiled'):
        dtype = guess_numpy_type(operator.inputs[0].type)
        fct = fct[FctVersion((dtype, ), None)]
    if hasattr(fct, 'compiled'):
        compiled = fct.compiled
        if not hasattr(compiled, 'onnx_'):
            raise RuntimeError(  # pragma: no cover
                "Attribute 'onnx_' is missing, function was not "
                "converted to onnx.")
        onx = compiled.onnx_
        graph = onx.graph
        nodes = graph.node

        # renaming all intermediate variables
        names = []
        for node in nodes:
            for name in node.input:
                names.append(name)
            for name in node.output:
                names.append(name)
        names = set(names)
        names_mapping = {}
        for name in names:
            names_mapping[name] = scope.get_unique_variable_name(
                'ft_%s' % name)

        # adding identities
        apply_identity(scope, operator.inputs[0].full_name,
                       names_mapping[graph.input[0].name], container)
        apply_identity(scope, names_mapping[graph.output[0].name],
                       operator.outputs[0].full_name, container)

        # adding initializers
        for init in graph.initializer:
            init = copy.deepcopy(init)
            name = names_mapping[init.name]
            init.name = name
            content = init.SerializeToString()
            container.initializers_strings[content] = name
            container.initializers.append(init)

        # adding nodes
        for node in nodes:
            atts = {}
            for att in node.attribute:
                atts[att.name] = _copy_attributes(att)
            container.add_node(
                node.op_type,
                [names_mapping[n] for n in node.input],
                [names_mapping[n] for n in node.output],
                name=scope.get_unique_operator_name('ft_%s' % node.op_type),
                **atts)
        return

    if op.func is not None:
        raise TypeError(  # pragma: no cover
            "FunctionTransformer is not supported unless the "
            "transform function is of type %r or "
            "wrapped with onnxnumpy." % type(op.func))
    if len(operator.inputs) == 1:
        apply_identity(scope, operator.inputs[0].full_name,
                       operator.outputs[0].full_name, container)
    else:
        apply_concat(scope, [i.full_name for i in operator.inputs],
                     operator.outputs[0].full_name, container)



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRunUnaryNum


class Softmax(OpRunUnaryNum):

    atts = {'axis': 1}

    def __init__(self, onnx_node, desc=None, **options):
        OpRunUnaryNum.__init__(self, onnx_node, desc=desc,
                               expected_attributes=Softmax.atts,
                               **options)

    def _run(self, X):  # pylint: disable=W0221
        if self.inplaces.get(0, False):
            return self._run_inplace(X)
        tmp = X - X.max(axis=self.axis)[:, numpy.newaxis]
        Y = numpy.exp(tmp)
        Y /= Y.sum(axis=self.axis)[:, numpy.newaxis]
        return (Y, )

    def _run_inplace(self, X):
        X -= X.max(axis=self.axis)[:, numpy.newaxis]
        numpy.exp(X, out=X)
        X /= X.sum(axis=self.axis)[:, numpy.newaxis]
        return (X, )

    def to_python(self, inputs):
        lines = ["tmp = {0} - {0}.max(axis=axis)[:, numpy.newaxis]".format(
            inputs[0]),
            "Y = numpy.exp(tmp)",
            "Y /= Y.sum(axis=axis)[:, numpy.newaxis]",
            "return Y"]
        return ("import numpy", "\n".join(lines))



"""
@file
@brief Helpers for notebooks.
"""
from IPython.core.magic import magics_class, line_magic
from jyquickhelper import RenderJsDot
from pyquickhelper.ipythonhelper import MagicCommandParser, MagicClassWithHelpers
from pyquickhelper.cli.cli_helper import create_cli_parser
from ..onnx_inference import OnnxInference


def onnxview(graph, recursive=False, local=False, add_rt_shapes=False,
             runtime='python', size=None):
    """
    Displays an :epkg:`ONNX` graph into a notebook.

    :param graph: filename, bytes, or :epkg:`onnx` graph.
    :param recursive: display subgraph
    :param local: use local path to javascript dependencies,
        recommanded option if used on :epkg:`MyBinder`)
    :param add_rt_shapes: add information about the shapes
        the runtime was able to find out,
        the runtime has to be `'python'`
    :param runtime: the view fails if a runtime does not implement a specific
        node unless *runtime* is `'empty'`
    :param size: graph size

    .. versionchanged:: 0.6
        Parameter *runtime* was added.
    """
    sess = OnnxInference(graph, skip_run=not add_rt_shapes, runtime=runtime)
    dot = sess.to_dot(recursive=recursive,
                      add_rt_shapes=add_rt_shapes, size=size)
    return RenderJsDot(dot, local=local)


@magics_class
class OnnxNotebook(MagicClassWithHelpers):

    """
    Defines magic commands to help with notebooks

    .. versionadded:: 1.1
    """

    @line_magic
    def onnxview(self, line):
        """
        Defines ``%onnxview``
        which displays an :epkg:`ONNX` graph.

        .. nbref::
            :title: onnxview

            The magic command ``%onnxview model_onnx`` is equivalent to function
            :func:`onnxview <mlprodict.onnxrt.doc.nb_helper.onnxview>`:

            ::

                onnx_view(model_onnx)

            It displays a visual representation of an :epkg:`ONNX` graph.

        """
        parser = self.get_parser(
            lambda: create_cli_parser(onnxview, cls=MagicCommandParser,
                                      positional={'graph'}), "onnxview")
        args = self.get_args(line, parser)

        if args is not None:
            res = onnxview(args.graph, recursive=args.recursive,
                           local=args.local, add_rt_shapes=args.add_rt_shapes)
            return res
        return None


def register_onnx_magics(ip=None):  # pragma: no cover
    """
    Register magics function, can be called from a notebook.

    @param      ip      from ``get_ipython()``
    """
    if ip is None:
        from IPython import get_ipython
        ip = get_ipython()
    ip.register_magics(OnnxNotebook)



"""
@file
@brief Types definition.
"""
import numpy
from .api_extension import AutoType


class MLType(AutoType):
    """
    Base class for every type.
    """

    def validate(self, value):
        """
        Checks that the value is of this type.
        """
        # It must be overwritten.
        self._cache = value

    def cast(self, value):
        """
        Converts *value* into this type.
        """
        raise NotImplementedError()  # pragma: no cover


class MLNumType(MLType):
    """
    Base class for numerical types.
    """

    def _format_value_json(self, value, hook=None):
        return str(value)

    def _format_value_c(self, value, hook=None):
        return str(value)

    def _copy_c(self, src, dst, hook=None):
        if hook == "typeref":
            return "*{0} = {1};".format(dst, src)
        return "{0} = {1};".format(dst, src)


class MLNumTypeSingle(MLNumType):
    """
    int32 or float32
    """

    def __init__(self, numpy_type, name, ctype, key):
        self.numpy_type = numpy_type
        self.name = name
        self.ctype = ctype
        self.key = key

    @property
    def CTypeSingle(self):
        """
        Returns *ctype*.
        """
        return self.ctype

    def validate(self, value):
        """
        Checks that the value is of this type.
        """
        MLNumType.validate(self, value)
        if not isinstance(value, self.numpy_type):
            raise TypeError(  # pragma: no cover
                "'{0}' is not a {1}.".format(
                    type(value), self.numpy_type))
        return value

    def cast(self, value):
        """
        Exports *value* into this type.
        """
        if isinstance(value, numpy.float32):
            raise TypeError(  # pragma: no cover
                "No need to cast, already a {0}".format(self.numpy_type))
        if isinstance(value, numpy.ndarray):
            if len(value) != 1:
                raise ValueError(  # pragma: no cover
                    "Dimension of array must be one single {0}".format(self.numpy_type))
            return value[0]
        raise NotImplementedError(  # pragma: no cover
            "Unable to cast '{0}' into a {0}".format(type(self.numpy_type)))

    def softcast(self, value):
        """
        Exports *value* into this type, does it anyway without verification.
        """
        if isinstance(value, numpy.ndarray):
            v = value.ravel()
            if len(v) != 1:
                raise ValueError(  # pragma: no cover
                    "Cannot cast shape {0} into {1}".format(
                        value.shape, self.numpy_type))
            return self.numpy_type(v[0])
        return self.numpy_type(value)

    def _export_common_c(self, ctype, hook=None, result_name=None):
        if hook == 'type':
            return {'code': ctype} if result_name is None else {'code': ctype + ' ' + result_name}
        if result_name is None:
            return {'code': ctype}
        return {'code': ctype + ' ' + result_name, 'result_name': result_name}

    def _byref_c(self):
        return "&"

    def _export_json(self, hook=None, result_name=None):
        return 'float32'

    def _export_c(self, hook=None, result_name=None):
        if hook == 'typeref':
            return {'code': self.ctype + '*'} if result_name is None else {'code': self.ctype + '* ' + result_name}
        return self._export_common_c(self.ctype, hook, result_name)

    def _format_value_json(self, value, hook=None):
        if hook is None or self.key not in hook:
            return value
        return hook[self.key](value)

    def _format_value_c(self, value, hook=None):
        if hook is None or self.key not in hook:
            return "({1}){0}".format(value, self.ctype)
        return hook[self.key](value)


class MLNumTypeFloat32(MLNumTypeSingle):
    """
    A numpy.float32.
    """

    def __init__(self):
        MLNumTypeSingle.__init__(
            self, numpy.float32, 'float32', 'float', 'float32')


class MLNumTypeFloat64(MLNumTypeSingle):
    """
    A numpy.float64.
    """

    def __init__(self):
        MLNumTypeSingle.__init__(
            self, numpy.float64, 'float64', 'double', 'float64')


class MLNumTypeInt32(MLNumTypeSingle):
    """
    A numpy.int32.
    """

    def __init__(self):
        MLNumTypeSingle.__init__(self, numpy.int32, 'int32', 'int', 'int32')


class MLNumTypeInt64(MLNumTypeSingle):
    """
    A numpy.int64.
    """

    def __init__(self):
        MLNumTypeSingle.__init__(
            self, numpy.int32, 'int64', 'int64_t', 'int64')


class MLNumTypeBool(MLNumTypeSingle):
    """
    A numpy.bool.
    """

    def __init__(self):
        MLNumTypeSingle.__init__(self, numpy.bool_, 'BL', 'bool', 'bool')


class MLTensor(MLType):
    """
    Defines a tensor with a dimension and a single type for what it contains.
    """

    def __init__(self, element_type, dim):
        if not isinstance(element_type, MLType):
            raise TypeError(  # pragma: no cover
                'element_type must be of MLType not {0}'.format(type(element_type)))
        if not isinstance(dim, tuple):
            raise TypeError(  # pragma: no cover
                'dim must be a tuple.')
        if len(dim) == 0:
            raise ValueError(  # pragma: no cover
                "dimension must not be null.")
        for d in dim:
            if d == 0:
                raise ValueError(  # pragma: no cover
                    "No dimension can be null.")
        self.dim = dim
        self.element_type = element_type

    @property
    def CTypeSingle(self):
        """
        Returns *ctype*.
        """
        return self.element_type.ctype

    def validate(self, value):
        """
        Checks that the value is of this type.
        """
        MLType.validate(self, value)
        if not isinstance(value, numpy.ndarray):
            raise TypeError(  # pragma: no cover
                "value is not a numpy.array but '{0}'".format(type(value)))
        if self.dim != value.shape:
            raise ValueError(  # pragma: no cover
                "Dimensions do not match {0}={1}".format(self.dim, value.shape))
        rvalue = value.ravel()
        for i, num in enumerate(rvalue):
            try:
                self.element_type.validate(num)
            except TypeError as e:  # pragma: no cover
                raise TypeError(
                    'Unable to convert an array due to value index {0}: {1}'.format(i, rvalue[i])) from e
        return value

    def _byref_c(self):
        return ""

    def _format_value_json(self, value, hook=None):
        if hook is None or 'array' not in hook:
            return value
        return hook['array'](value)

    def _format_value_c(self, value, hook=None):
        return "{{{0}}}".format(", ".join(self.element_type._format_value_c(x) for x in value))

    def _export_json(self, hook=None, result_name=None):
        return '{0}:{1}'.format(self.element_type._export_json(hook=hook), self.dim)

    def _export_c(self, hook=None, result_name=None):
        if len(self.dim) != 1:
            raise NotImplementedError(  # pragma: no cover
                'Only 1D vector implemented.')
        if hook is None:
            raise ValueError(  # pragma: no cover
                "hook must contains either 'signature' or 'declare'.")
        if hook == 'signature':
            if result_name is None:
                raise ValueError(  # pragma: no cover
                    "result_name must be specified.")
            return {'code': "{0}[{1}] {2}".format(self.element_type._export_c(hook=hook)['code'],
                                                  self.dim[0], result_name),
                    'result_name': result_name}
        elif hook == 'declare':
            if result_name is None:
                raise ValueError(  # pragma: no cover
                    "result_name must be specified.")
            dc = self.element_type._export_c(
                hook=hook, result_name=result_name)
            return {'code': "{0}[{1}]".format(dc['code'], self.dim[0])}
        elif hook == 'type':
            return {'code': "{0}*".format(self.element_type._export_c(hook=hook)['code'])}
        elif hook == 'typeref':
            if result_name is None:
                return {'code': "{0}*".format(self.element_type._export_c(hook='type')['code'])}
            code = self.element_type._export_c(hook='type')['code']
            return {'code': "{0}* {1}".format(code, result_name), 'result_name': result_name}
        else:
            raise ValueError(  # pragma: no cover
                "hook must contains either 'signature' or 'declare' not '{0}'.".format(hook))

    def _copy_c(self, src, dest, hook=None):
        if len(self.dim) != 1:
            raise NotImplementedError(  # pragma: no cover
                'Only 1D vector implemented.')
        code = self.element_type._export_c(hook='type')['code']
        return "memcpy({1}, {0}, {2}*sizeof({3}));".format(src, dest, self.dim[0], code)



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
from ._op import OpRun


class If(OpRun):

    atts = {
        'then_branch': None,
        'else_branch': None,
    }

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       expected_attributes=If.atts,
                       **options)
        if not hasattr(self.then_branch, 'run'):
            raise RuntimeError(  # pragma: no cover
                "Parameter 'then_branch' must have a method 'run', "
                "type {}.".format(type(self.then_branch)))
        if not hasattr(self.else_branch, 'run'):
            raise RuntimeError(  # pragma: no cover
                "Parameter 'else_branch' must have a method 'run', "
                "type {}.".format(type(self.else_branch)))

        self._run_meth_then = (self.then_branch.run_in_scan
                               if hasattr(self.then_branch, 'run_in_scan')
                               else self.then_branch.run)
        self._run_meth_else = (self.else_branch.run_in_scan
                               if hasattr(self.else_branch, 'run_in_scan')
                               else self.else_branch.run)

    def _run(self, cond, named_inputs=None):  # pylint: disable=W0221
        if named_inputs is None:
            named_inputs = {}
        if len(self.then_branch.input_names) > 0:
            if len(named_inputs) == 0:
                raise RuntimeError(  # pragma: no cover
                    "named_inputs is empty but the graph needs {}.".format(
                        self.then_branch.input_names))
            for k in self.then_branch.input_names:
                if k not in named_inputs:
                    raise RuntimeError(  # pragma: no cover
                        "Unable to find named input '{}' in\n{}.".format(
                            k, "\n".join(sorted(named_inputs))))
        if len(self.else_branch.input_names) > 0:
            if len(named_inputs) == 0:
                raise RuntimeError(  # pragma: no cover
                    "named_inputs is empty but the graph needs {}.".format(
                        self.then_branch.input_names))
            for k in self.else_branch.input_names:
                if k not in named_inputs:
                    raise RuntimeError(  # pragma: no cover
                        "Unable to find named input '{}' in\n{}.".format(
                            k, "\n".join(sorted(named_inputs))))

        if all(cond):
            outputs = self._run_meth_then(named_inputs)
            return tuple([outputs[name] for name in self.then_branch.output_names])
        outputs = self._run_meth_else(named_inputs)
        return tuple([outputs[name] for name in self.else_branch.output_names])

    def _infer_shapes(self, cond, named_inputs=None):  # pylint: disable=W0221
        res = self.then_branch._set_shape_inference_runtime()
        return tuple([res[name] for name in self.then_branch.output_names])

    def _infer_types(self, cond, named_inputs=None):  # pylint: disable=W0221
        res = self.then_branch._set_type_inference_runtime()
        return tuple([res[name] for name in self.then_branch.output_names])



# -*- encoding: utf-8 -*-
"""
@file
@brief Ways to speed up predictions for a machine learned model.
"""

__version__ = "0.6.1522"
__author__ = "Xavier Dupré"


def check(log=False):
    """
    Checks the library is working.
    It raises an exception.
    If you want to disable the logs:

    @param      log     if True, display information, otherwise
    @return             0 or exception
    """
    return True


def _setup_hook(use_print=False):
    """
    If this function is added to the module,
    the help automation and unit tests call it first before
    anything goes on as an initialization step.
    """
    if use_print:
        print("Success: _setup_hook")


def load_ipython_extension(ip):  # pragma: no cover
    """
    To allow the call ``%load_ext mlprodict``

    @param      ip      from ``get_ipython()``
    """
    from .onnxrt.doc.nb_helper import register_onnx_magics as freg
    freg(ip)



"""
@file
@brief :epkg:`numpy` functions implemented with :epkg:`onnx`
and compiled with this python runtime.

.. versionadded:: 0.6
"""
import numpy
from .onnx_numpy_annotation import NDArrayType
from .numpy_onnx_impl_skl import (
    logistic_regression as nx_logistic_regression,
)
from .onnx_numpy_wrapper import onnxnumpy_np


@onnxnumpy_np(signature=NDArrayType(("T:all", ), dtypes_out=((numpy.int64,), "T")))
def logistic_regression(x, *, model=None):
    "logistic_regression"
    return nx_logistic_regression(x, model=model)



"""
@file
@brief Validates runtime for many :scikit-learn: operators.
The submodule relies on :epkg:`onnxconverter_common`,
:epkg:`sklearn-onnx`.
"""
import pprint
from inspect import signature
import numpy
from numpy.linalg import LinAlgError
import sklearn
from sklearn import __all__ as sklearn__all__, __version__ as sklearn_version
from sklearn.exceptions import ConvergenceWarning
from sklearn.utils._testing import ignore_warnings
from ... import __version__ as ort_version
from ...onnx_conv import to_onnx, register_converters, register_rewritten_operators
from ...tools.ort_wrapper import onnxrt_version
from ...tools.model_info import analyze_model, set_random_state
from ...tools.asv_options_helper import (
    get_opset_number_from_onnx, get_ir_version_from_onnx)
from ..onnx_inference import OnnxInference
from ...onnx_tools.optim.sklearn_helper import inspect_sklearn_model, set_n_jobs
from ...onnx_tools.optim.onnx_helper import onnx_statistics
from ...onnx_tools.optim import onnx_optimisations
from .validate_problems import find_suitable_problem
from .validate_scenarios import _extra_parameters
from .validate_difference import measure_relative_difference
from .validate_helper import (
    _dispsimple, sklearn_operators,
    _measure_time, _shape_exc, dump_into_folder,
    default_time_kwargs, RuntimeBadResultsError,
    _dictionary2str, _merge_options, _multiply_time_kwargs,
    _get_problem_data)
from .validate_benchmark import benchmark_fct


@ignore_warnings(category=(UserWarning, ConvergenceWarning))
def _dofit_model(dofit, obs, inst, X_train, y_train, X_test, y_test,
                 Xort_test, init_types, store_models,
                 debug, verbose, fLOG):
    if dofit:
        if verbose >= 2 and fLOG is not None:
            fLOG("[enumerate_compatible_opset] fit, type: '{}' dtype: {}".format(
                type(X_train), getattr(X_train, 'dtype', '-')))
        try:
            set_random_state(inst)
            if y_train is None:
                t4 = _measure_time(lambda: inst.fit(X_train))[1]
            else:
                t4 = _measure_time(
                    lambda: inst.fit(X_train, y_train))[1]
        except (AttributeError, TypeError, ValueError,
                IndexError, NotImplementedError, MemoryError,
                LinAlgError, StopIteration) as e:
            if debug:
                raise  # pragma: no cover
            obs["_1training_time_exc"] = str(e)
            return False

        obs["training_time"] = t4
        try:
            skl_st = inspect_sklearn_model(inst)
        except NotImplementedError:
            skl_st = {}
        obs.update({'skl_' + k: v for k, v in skl_st.items()})

        if store_models:
            obs['MODEL'] = inst
            obs['X_test'] = X_test
            obs['Xort_test'] = Xort_test
            obs['init_types'] = init_types
    else:
        obs["training_time"] = 0.
        if store_models:
            obs['MODEL'] = inst
            obs['init_types'] = init_types

    return True


def _run_skl_prediction(obs, check_runtime, assume_finite, inst,
                        method_name, predict_kwargs, X_test,
                        benchmark, debug, verbose, time_kwargs,
                        skip_long_test, time_kwargs_fact, fLOG):
    if not check_runtime:
        return None  # pragma: no cover
    if verbose >= 2 and fLOG is not None:
        fLOG("[enumerate_compatible_opset] check_runtime SKL {}-{}-{}-{}-{}".format(
            id(inst), method_name, predict_kwargs, time_kwargs,
            time_kwargs_fact))
    with sklearn.config_context(assume_finite=assume_finite):
        # compute sklearn prediction
        obs['ort_version'] = ort_version
        try:
            meth = getattr(inst, method_name)
        except AttributeError as e:
            if debug:
                raise  # pragma: no cover
            obs['_2skl_meth_exc'] = str(e)
            return e
        try:
            ypred, t4, ___ = _measure_time(
                lambda: meth(X_test, **predict_kwargs))
            obs['lambda-skl'] = (lambda xo: meth(xo, **predict_kwargs), X_test)
        except (ValueError, AttributeError, TypeError, MemoryError, IndexError) as e:
            if debug:
                raise  # pragma: no cover
            obs['_3prediction_exc'] = str(e)
            return e
        obs['prediction_time'] = t4
        obs['assume_finite'] = assume_finite
        if benchmark and 'lambda-skl' in obs:
            obs['bench-skl'] = benchmark_fct(
                *obs['lambda-skl'], obs=obs,
                time_kwargs=_multiply_time_kwargs(
                    time_kwargs, time_kwargs_fact, inst),
                skip_long_test=skip_long_test)
        if verbose >= 3 and fLOG is not None:
            fLOG("[enumerate_compatible_opset] scikit-learn prediction")
            _dispsimple(ypred, fLOG)
        if verbose >= 2 and fLOG is not None:
            fLOG("[enumerate_compatible_opset] predictions stored")
    return ypred


def _retrieve_problems_extra(model, verbose, fLOG, extended_list):
    """
    Use by @see fn enumerate_compatible_opset.
    """
    extras = None
    if extended_list:
        from ...onnx_conv.validate_scenarios import find_suitable_problem as fsp_extended
        problems = fsp_extended(model)
        if problems is not None:
            from ...onnx_conv.validate_scenarios import build_custom_scenarios as fsp_scenarios
            extra_parameters = fsp_scenarios()

            if verbose >= 2 and fLOG is not None:
                fLOG(
                    "[enumerate_compatible_opset] found custom for model={}".format(model))
                extras = extra_parameters.get(model, None)
                if extras is not None:
                    fLOG(
                        "[enumerate_compatible_opset] found custom scenarios={}".format(extras))
    else:
        problems = None

    if problems is None:
        # scikit-learn
        extra_parameters = _extra_parameters
        try:
            problems = find_suitable_problem(model)
        except RuntimeError as e:
            return {'name': model.__name__, 'skl_version': sklearn_version,
                    '_0problem_exc': e}, extras
    extras = extra_parameters.get(model, [('default', {})])

    # checks existence of random_state
    sig = signature(model.__init__)
    if 'random_state' in sig.parameters:
        new_extras = []
        for extra in extras:
            if 'random_state' not in extra[1]:
                ps = extra[1].copy()
                ps['random_state'] = 42
                if len(extra) == 2:
                    extra = (extra[0], ps)
                else:
                    extra = (extra[0], ps) + extra[2:]
            new_extras.append(extra)
        extras = new_extras

    return problems, extras


def enumerate_compatible_opset(model, opset_min=-1, opset_max=-1,  # pylint: disable=R0914
                               check_runtime=True, debug=False,
                               runtime='python', dump_folder=None,
                               store_models=False, benchmark=False,
                               assume_finite=True, node_time=False,
                               fLOG=print, filter_exp=None,
                               verbose=0, time_kwargs=None,
                               extended_list=False, dump_all=False,
                               n_features=None, skip_long_test=True,
                               filter_scenario=None, time_kwargs_fact=None,
                               time_limit=4, n_jobs=None):
    """
    Lists all compatible opsets for a specific model.

    @param      model           operator class
    @param      opset_min       starts with this opset
    @param      opset_max       ends with this opset (None to use
                                current onnx opset)
    @param      check_runtime   checks that runtime can consume the
                                model and compute predictions
    @param      debug           catch exception (True) or not (False)
    @param      runtime         test a specific runtime, by default ``'python'``
    @param      dump_folder     dump information to replicate in case of mismatch
    @param      dump_all        dump all models not only the one which fail
    @param      store_models    if True, the function
                                also stores the fitted model and its conversion
                                into :epkg:`ONNX`
    @param      benchmark       if True, measures the time taken by each function
                                to predict for different number of rows
    @param      fLOG            logging function
    @param      filter_exp      function which tells if the experiment must be run,
                                None to run all, takes *model, problem* as an input
    @param      filter_scenario second function which tells if the experiment must be run,
                                None to run all, takes *model, problem, scenario, extra, options*
                                as an input
    @param      node_time       collect time for each node in the :epkg:`ONNX` graph
    @param      assume_finite   See `config_context
                                <https://scikit-learn.org/stable/modules/generated/
                                sklearn.config_context.html>`_, If True, validation for finiteness
                                will be skipped, saving time, but leading to potential crashes.
                                If False, validation for finiteness will be performed, avoiding error.
    @param      verbose         verbosity
    @param      extended_list   extends the list to custom converters
                                and problems
    @param      time_kwargs     to define a more precise way to measure a model
    @param      n_features      modifies the shorts datasets used to train the models
                                to use exactly this number of features, it can also
                                be a list to test multiple datasets
    @param      skip_long_test  skips tests for high values of N if they seem too long
    @param      time_kwargs_fact see :func:`_multiply_time_kwargs <mlprodict.onnxrt.validate.validate_helper._multiply_time_kwargs>`
    @param      time_limit      to stop benchmarking after this amount of time was spent
    @param      n_jobs          *n_jobs* is set to the number of CPU by default unless this
                                value is changed
    @return                     dictionaries, each row has the following
                                keys: opset, exception if any, conversion time,
                                problem chosen to test the conversion...

    The function requires :epkg:`sklearn-onnx`.
    The outcome can be seen at pages references
    by :ref:`l-onnx-availability`.
    The parameter *time_kwargs* is a dictionary which defines the
    number of times to repeat the same predictions in order
    to give more precise figures. The default value (if None) is returned
    by the following code:

    .. runpython::
        :showcode:
        :warningout: DeprecationWarning

        from mlprodict.onnxrt.validate.validate_helper import default_time_kwargs
        import pprint
        pprint.pprint(default_time_kwargs())

    Parameter *time_kwargs_fact* multiples these values for some
    specific models. ``'lin'`` multiplies by 10 when the model
    is linear.
    """
    if opset_min == -1:
        opset_min = get_opset_number_from_onnx()  # pragma: no cover
    if opset_max == -1:
        opset_max = get_opset_number_from_onnx()  # pragma: no cover
    if verbose > 0 and fLOG is not None:
        fLOG("[enumerate_compatible_opset] opset in [{}, {}].".format(
            opset_min, opset_max))
    if verbose > 1 and fLOG:
        fLOG("[enumerate_compatible_opset] validate class '{}'.".format(
            model.__name__))
        if verbose > 2:
            fLOG(model)

    if time_kwargs is None:
        time_kwargs = default_time_kwargs()
    problems, extras = _retrieve_problems_extra(
        model, verbose, fLOG, extended_list)
    if isinstance(problems, dict):
        yield problems  # pragma: no cover
        problems = []  # pragma: no cover

    if opset_max is None:
        opset_max = get_opset_number_from_onnx()  # pragma: no cover
        opsets = list(range(opset_min, opset_max + 1))  # pragma: no cover
        opsets.append(None)  # pragma: no cover
    else:
        opsets = list(range(opset_min, opset_max + 1))

    if extras is None:
        problems = []
        yield {'name': model.__name__, 'skl_version': sklearn_version,
               '_0problem_exc': 'SKIPPED'}

    if not isinstance(n_features, list):
        n_features = [n_features]

    for prob in problems:
        if filter_exp is not None and not filter_exp(model, prob):
            continue
        for n_feature in n_features:
            if verbose >= 2 and fLOG is not None:
                fLOG("[enumerate_compatible_opset] problem={} n_feature={}".format(
                    prob, n_feature))

            (X_train, X_test, y_train,
             y_test, Xort_test,
             init_types, conv_options, method_name,
             output_index, dofit, predict_kwargs) = _get_problem_data(prob, n_feature)

            for scenario_extra in extras:
                subset_problems = None
                optimisations = None
                new_conv_options = None
                if len(scenario_extra) > 2:
                    options = scenario_extra[2]
                    if isinstance(options, dict):
                        subset_problems = options.get('subset_problems', None)
                        optimisations = options.get('optim', None)
                        new_conv_options = options.get('conv_options', None)
                    else:
                        subset_problems = options

                if subset_problems and isinstance(subset_problems, (list, set)):
                    if prob not in subset_problems:
                        # Skips unrelated problem for a specific configuration.
                        continue
                elif subset_problems is not None:
                    raise RuntimeError(  # pragma: no cover
                        "subset_problems must be a set or a list not {}.".format(
                            subset_problems))

                try:
                    scenario, extra = scenario_extra[:2]
                except TypeError as e:  # pragma: no cover
                    raise TypeError(
                        "Unable to interpret 'scenario_extra'\n{}".format(
                            scenario_extra)) from e
                if optimisations is None:
                    optimisations = [None]
                if new_conv_options is None:
                    new_conv_options = [{}]

                if (filter_scenario is not None and
                        not filter_scenario(model, prob, scenario,
                                            extra, new_conv_options)):
                    continue

                if verbose >= 2 and fLOG is not None:
                    fLOG("[enumerate_compatible_opset] ##############################")
                    fLOG("[enumerate_compatible_opset] scenario={} optim={} extra={} dofit={} (problem={})".format(
                        scenario, optimisations, extra, dofit, prob))

                # training
                obs = {'scenario': scenario, 'name': model.__name__,
                       'skl_version': sklearn_version, 'problem': prob,
                       'method_name': method_name, 'output_index': output_index,
                       'fit': dofit, 'conv_options': conv_options,
                       'idtype': Xort_test.dtype, 'predict_kwargs': predict_kwargs,
                       'init_types': init_types, 'inst': extra if extra else None,
                       'n_features': X_train.shape[1] if len(X_train.shape) == 2 else 1}
                inst = None
                extra = set_n_jobs(model, extra, n_jobs=n_jobs)
                try:
                    inst = model(**extra)
                except TypeError as e:  # pragma: no cover
                    if debug:  # pragma: no cover
                        raise
                    if "__init__() missing" not in str(e):
                        raise RuntimeError(
                            "Unable to instantiate model '{}'.\nextra=\n{}".format(
                                model.__name__, pprint.pformat(extra))) from e
                    yield obs.copy()
                    continue

                if not _dofit_model(dofit, obs, inst, X_train, y_train, X_test, y_test,
                                    Xort_test, init_types, store_models,
                                    debug, verbose, fLOG):
                    yield obs.copy()
                    continue

                # statistics about the trained model
                skl_infos = analyze_model(inst)
                for k, v in skl_infos.items():
                    obs['fit_' + k] = v

                # runtime
                ypred = _run_skl_prediction(
                    obs, check_runtime, assume_finite, inst,
                    method_name, predict_kwargs, X_test,
                    benchmark, debug, verbose, time_kwargs,
                    skip_long_test, time_kwargs_fact, fLOG)
                if isinstance(ypred, Exception):
                    yield obs.copy()
                    continue

                for run_obs in _call_conv_runtime_opset(
                        obs=obs.copy(), opsets=opsets, debug=debug,
                        new_conv_options=new_conv_options,
                        model=model, prob=prob, scenario=scenario,
                        extra=extra, extras=extras, conv_options=conv_options,
                        init_types=init_types, inst=inst,
                        optimisations=optimisations, verbose=verbose,
                        benchmark=benchmark, runtime=runtime,
                        filter_scenario=filter_scenario,
                        X_test=X_test, y_test=y_test, ypred=ypred,
                        Xort_test=Xort_test, method_name=method_name,
                        check_runtime=check_runtime,
                        output_index=output_index,
                        kwargs=dict(
                            dump_all=dump_all,
                            dump_folder=dump_folder,
                            node_time=node_time,
                            skip_long_test=skip_long_test,
                            store_models=store_models,
                            time_kwargs=_multiply_time_kwargs(
                                time_kwargs, time_kwargs_fact, inst)
                        ),
                        time_limit=time_limit,
                        fLOG=fLOG):
                    yield run_obs


def _check_run_benchmark(benchmark, stat_onnx, bench_memo, runtime):
    unique = set(stat_onnx.items())
    unique.add(runtime)
    run_benchmark = benchmark and all(
        map(lambda u: unique != u, bench_memo))
    if run_benchmark:
        bench_memo.append(unique)
    return run_benchmark


def _call_conv_runtime_opset(
        obs, opsets, debug, new_conv_options,
        model, prob, scenario, extra, extras, conv_options,
        init_types, inst, optimisations, verbose,
        benchmark, runtime, filter_scenario,
        check_runtime, X_test, y_test, ypred, Xort_test,
        method_name, output_index,
        kwargs, time_limit, fLOG):
    # Calls the conversion and runtime for different opets
    if None in opsets:
        set_opsets = [None] + list(sorted((_ for _ in opsets if _ is not None),
                                          reverse=True))
    else:
        set_opsets = list(sorted(opsets, reverse=True))
    bench_memo = []

    for opset in set_opsets:
        if verbose >= 2 and fLOG is not None:
            fLOG("[enumerate_compatible_opset] opset={} init_types={}".format(
                opset, init_types))
        obs_op = obs.copy()
        if opset is not None:
            obs_op['opset'] = opset

        if len(init_types) != 1:
            raise NotImplementedError(  # pragma: no cover
                "Multiple types are is not implemented: "
                "{}.".format(init_types))

        if not isinstance(runtime, list):
            runtime = [runtime]

        obs_op_0c = obs_op.copy()
        for aoptions in new_conv_options:
            obs_op = obs_op_0c.copy()
            all_conv_options = {} if conv_options is None else conv_options.copy()
            all_conv_options = _merge_options(
                all_conv_options, aoptions)
            obs_op['conv_options'] = all_conv_options

            if (filter_scenario is not None and
                    not filter_scenario(model, prob, scenario,
                                        extra, all_conv_options)):
                continue

            for rt in runtime:
                def fct_conv(itt=inst, it=init_types[0][1], ops=opset,
                             options=all_conv_options):
                    return to_onnx(itt, it, target_opset=ops, options=options,
                                   rewrite_ops=rt in ('', None, 'python',
                                                      'python_compiled'))

                if verbose >= 2 and fLOG is not None:
                    fLOG(
                        "[enumerate_compatible_opset] conversion to onnx: {}".format(all_conv_options))
                try:
                    conv, t4 = _measure_time(fct_conv)[:2]
                    obs_op["convert_time"] = t4
                except (RuntimeError, IndexError, AttributeError, TypeError,
                        ValueError, NameError, NotImplementedError) as e:
                    if debug:
                        fLOG(pprint.pformat(obs_op))  # pragma: no cover
                        raise  # pragma: no cover
                    obs_op["_4convert_exc"] = e
                    yield obs_op.copy()
                    continue

                if verbose >= 6 and fLOG is not None:
                    fLOG(  # pragma: no cover
                        "[enumerate_compatible_opset] ONNX:\n{}".format(conv))

                if all_conv_options.get('optim', '') == 'cdist':  # pragma: no cover
                    check_cdist = [_ for _ in str(conv).split('\n')
                                   if 'CDist' in _]
                    check_scan = [_ for _ in str(conv).split('\n')
                                  if 'Scan' in _]
                    if len(check_cdist) == 0 and len(check_scan) > 0:
                        raise RuntimeError(
                            "Operator CDist was not used in\n{}"
                            "".format(conv))

                obs_op0 = obs_op.copy()
                for optimisation in optimisations:
                    obs_op = obs_op0.copy()
                    if optimisation is not None:
                        if optimisation == 'onnx':
                            obs_op['optim'] = optimisation
                            if len(aoptions) != 0:
                                obs_op['optim'] += '/' + \
                                    _dictionary2str(aoptions)
                            conv = onnx_optimisations(conv)
                        else:
                            raise ValueError(  # pragma: no cover
                                "Unknown optimisation option '{}' (extra={})"
                                "".format(optimisation, extras))
                    else:
                        obs_op['optim'] = _dictionary2str(aoptions)

                    if verbose >= 3 and fLOG is not None:
                        fLOG("[enumerate_compatible_opset] optim='{}' optimisation={} all_conv_options={}".format(
                            obs_op['optim'], optimisation, all_conv_options))
                    if kwargs['store_models']:
                        obs_op['ONNX'] = conv
                        if verbose >= 2 and fLOG is not None:
                            fLOG(  # pragma: no cover
                                "[enumerate_compatible_opset] onnx nodes: {}".format(
                                    len(conv.graph.node)))
                    stat_onnx = onnx_statistics(conv)
                    obs_op.update(
                        {'onx_' + k: v for k, v in stat_onnx.items()})

                    # opset_domain
                    for op_imp in list(conv.opset_import):
                        obs_op['domain_opset_%s' %
                               op_imp.domain] = op_imp.version

                    run_benchmark = _check_run_benchmark(
                        benchmark, stat_onnx, bench_memo, rt)

                    # prediction
                    if check_runtime:
                        yield _call_runtime(obs_op=obs_op.copy(), conv=conv,
                                            opset=opset, debug=debug,
                                            runtime=rt, inst=inst,
                                            X_test=X_test, y_test=y_test,
                                            init_types=init_types,
                                            method_name=method_name,
                                            output_index=output_index,
                                            ypred=ypred, Xort_test=Xort_test,
                                            model=model,
                                            dump_folder=kwargs['dump_folder'],
                                            benchmark=run_benchmark,
                                            node_time=kwargs['node_time'],
                                            time_kwargs=kwargs['time_kwargs'],
                                            fLOG=fLOG, verbose=verbose,
                                            store_models=kwargs['store_models'],
                                            dump_all=kwargs['dump_all'],
                                            skip_long_test=kwargs['skip_long_test'],
                                            time_limit=time_limit)
                    else:
                        yield obs_op.copy()  # pragma: no cover


def _call_runtime(obs_op, conv, opset, debug, inst, runtime,
                  X_test, y_test, init_types, method_name, output_index,
                  ypred, Xort_test, model, dump_folder,
                  benchmark, node_time, fLOG,
                  verbose, store_models, time_kwargs,
                  dump_all, skip_long_test, time_limit):
    """
    Private.
    """
    if 'onnxruntime' in runtime:
        old = conv.ir_version
        conv.ir_version = get_ir_version_from_onnx()
    else:
        old = None

    ser, t5, ___ = _measure_time(lambda: conv.SerializeToString())
    obs_op['tostring_time'] = t5
    obs_op['runtime'] = runtime

    if old is not None:
        conv.ir_version = old

    # load
    if verbose >= 2 and fLOG is not None:
        fLOG("[enumerate_compatible_opset-R] load onnx")
    try:
        sess, t5, ___ = _measure_time(
            lambda: OnnxInference(ser, runtime=runtime))
        obs_op['tostring_time'] = t5
    except (RuntimeError, ValueError, KeyError, IndexError, TypeError) as e:
        if debug:
            raise  # pragma: no cover
        obs_op['_5ort_load_exc'] = e
        return obs_op

    # compute batch
    if store_models:
        obs_op['OINF'] = sess
    if verbose >= 2 and fLOG is not None:
        fLOG("[enumerate_compatible_opset-R] compute batch with runtime "
             "'{}'".format(runtime))

    def fct_batch(se=sess, xo=Xort_test, it=init_types):  # pylint: disable=W0102
        return se.run({it[0][0]: xo},
                      verbose=max(verbose - 1, 1) if debug else 0, fLOG=fLOG)

    try:
        opred, t5, ___ = _measure_time(fct_batch)
        obs_op['ort_run_time_batch'] = t5
        obs_op['lambda-batch'] = (lambda xo: sess.run(
            {init_types[0][0]: xo}, node_time=node_time), Xort_test)
    except (RuntimeError, TypeError, ValueError, KeyError, IndexError) as e:
        if debug:
            raise RuntimeError("Issue with {}.".format(
                obs_op)) from e  # pragma: no cover
        obs_op['_6ort_run_batch_exc'] = e
    if (benchmark or node_time) and 'lambda-batch' in obs_op:
        try:
            benres = benchmark_fct(*obs_op['lambda-batch'], obs=obs_op,
                                   node_time=node_time, time_kwargs=time_kwargs,
                                   skip_long_test=skip_long_test,
                                   time_limit=time_limit)
            obs_op['bench-batch'] = benres
        except (RuntimeError, TypeError, ValueError) as e:  # pragma: no cover
            if debug:
                raise e  # pragma: no cover
            obs_op['_6ort_run_batch_exc'] = e
            obs_op['_6ort_run_batch_bench_exc'] = e

    # difference
    debug_exc = []
    if verbose >= 2 and fLOG is not None:
        fLOG("[enumerate_compatible_opset-R] differences")
    if '_6ort_run_batch_exc' not in obs_op:
        if isinstance(opred, dict):
            ch = [(k, v) for k, v in opred.items()]
            opred = [_[1] for _ in ch]

        if output_index != 'all':
            try:
                opred = opred[output_index]
            except IndexError as e:  # pragma: no cover
                if debug:
                    raise IndexError(
                        "Issue with output_index={}/{}".format(
                            output_index, len(opred))) from e
                obs_op['_8max_rel_diff_batch_exc'] = (
                    "Unable to fetch output {}/{} for model '{}'"
                    "".format(output_index, len(opred),
                              model.__name__))
                opred = None

        if opred is not None:
            if store_models:
                obs_op['skl_outputs'] = ypred
                obs_op['ort_outputs'] = opred
            if verbose >= 3 and fLOG is not None:
                fLOG("[_call_runtime] runtime prediction")
                _dispsimple(opred, fLOG)

            if (method_name == "decision_function" and hasattr(opred, 'shape') and
                    hasattr(ypred, 'shape') and len(opred.shape) == 2 and
                    opred.shape[1] == 2 and len(ypred.shape) == 1):
                # decision_function, for binary classification,
                # raw score is a distance
                max_rel_diff = measure_relative_difference(
                    ypred, opred[:, 1])
            else:
                max_rel_diff = measure_relative_difference(
                    ypred, opred)

            if max_rel_diff >= 1e9 and debug:  # pragma: no cover
                _shape = lambda o: o.shape if hasattr(
                    o, 'shape') else 'no shape'
                raise RuntimeError(
                    "Big difference (opset={}, runtime='{}' p='{}' s='{}')"
                    ":\n-------\n{}-{}\n{}\n--------\n{}-{}\n{}".format(
                        opset, runtime, obs_op['problem'], obs_op['scenario'],
                        type(ypred), _shape(ypred), ypred,
                        type(opred), _shape(opred), opred))

            if numpy.isnan(max_rel_diff):
                obs_op['_8max_rel_diff_batch_exc'] = (  # pragma: no cover
                    "Unable to compute differences between"
                    " {}-{}\n{}\n--------\n{}".format(
                        _shape_exc(
                            ypred), _shape_exc(opred),
                        ypred, opred))
                if debug:  # pragma: no cover
                    debug_exc.append(RuntimeError(
                        obs_op['_8max_rel_diff_batch_exc']))
            else:
                obs_op['max_rel_diff_batch'] = max_rel_diff
                if dump_folder and max_rel_diff > 1e-5:
                    dump_into_folder(dump_folder, kind='batch', obs_op=obs_op,
                                     X_test=X_test, y_test=y_test, Xort_test=Xort_test)
                if debug and max_rel_diff >= 0.1:  # pragma: no cover
                    raise RuntimeError("Two big differences {}\n{}\n{}\n{}".format(
                        max_rel_diff, inst, conv, pprint.pformat(obs_op)))

    if debug and len(debug_exc) == 2:
        raise debug_exc[0]  # pragma: no cover
    if debug and verbose >= 2:  # pragma: no cover
        if verbose >= 3:
            fLOG(pprint.pformat(obs_op))
        else:
            obs_op_log = {k: v for k,
                          v in obs_op.items() if 'lambda-' not in k}
            fLOG(pprint.pformat(obs_op_log))
    if verbose >= 2 and fLOG is not None:
        fLOG("[enumerate_compatible_opset-R] next...")
    if dump_all:
        dump = dump_into_folder(dump_folder, kind='batch', obs_op=obs_op,
                                X_test=X_test, y_test=y_test, Xort_test=Xort_test,
                                is_error=len(debug_exc) > 1,
                                onnx_bytes=conv.SerializeToString(),
                                skl_model=inst, ypred=ypred)
        obs_op['dumped'] = dump
    return obs_op


def _enumerate_validated_operator_opsets_ops(extended_list, models, skip_models):
    ops = [_ for _ in sklearn_operators(extended=extended_list)]

    if models is not None:
        if not all(map(lambda m: isinstance(m, str), models)):
            raise ValueError(  # pragma: no cover
                "models must be a set of strings.")
        ops_ = [_ for _ in ops if _['name'] in models]
        if len(ops) == 0:
            raise ValueError(  # pragma: no cover
                "Parameter models is wrong: {}\n{}".format(
                    models, ops[0]))
        ops = ops_
    if skip_models is not None:
        ops = [m for m in ops if m['name'] not in skip_models]
    return ops


def _enumerate_validated_operator_opsets_version(runtime):
    from numpy import __version__ as numpy_version
    from onnx import __version__ as onnx_version
    from scipy import __version__ as scipy_version
    from skl2onnx import __version__ as skl2onnx_version
    add_versions = {'v_numpy': numpy_version, 'v_onnx': onnx_version,
                    'v_scipy': scipy_version, 'v_skl2onnx': skl2onnx_version,
                    'v_sklearn': sklearn_version, 'v_onnxruntime': ort_version}
    if "onnxruntime" in runtime:
        add_versions['v_onnxruntime'] = onnxrt_version
    return add_versions


def enumerate_validated_operator_opsets(verbose=0, opset_min=-1, opset_max=-1,
                                        check_runtime=True, debug=False, runtime='python',
                                        models=None, dump_folder=None, store_models=False,
                                        benchmark=False, skip_models=None,
                                        assume_finite=True, node_time=False,
                                        fLOG=print, filter_exp=None,
                                        versions=False, extended_list=False,
                                        time_kwargs=None, dump_all=False,
                                        n_features=None, skip_long_test=True,
                                        fail_bad_results=False,
                                        filter_scenario=None,
                                        time_kwargs_fact=None,
                                        time_limit=4, n_jobs=None):
    """
    Tests all possible configurations for all possible
    operators and returns the results.

    @param      verbose         integer 0, 1, 2
    @param      opset_min       checks conversion starting from the opset, -1
                                to get the last one
    @param      opset_max       checks conversion up to this opset,
                                None means @see fn get_opset_number_from_onnx.
    @param      check_runtime   checks the python runtime
    @param      models          only process a small list of operators,
                                set of model names
    @param      debug           stops whenever an exception
                                is raised
    @param      runtime         test a specific runtime, by default ``'python'``
    @param      dump_folder     dump information to replicate in case of mismatch
    @param      dump_all        dump all models not only the one which fail
    @param      store_models    if True, the function
                                also stores the fitted model and its conversion
                                into :epkg:`ONNX`
    @param      benchmark       if True, measures the time taken by each function
                                to predict for different number of rows
    @param      filter_exp      function which tells if the experiment must be run,
                                None to run all, takes *model, problem* as an input
    @param      filter_scenario second function which tells if the experiment must be run,
                                None to run all, takes *model, problem, scenario, extra, options*
                                as an input
    @param      skip_models     models to skip
    @param      assume_finite   See `config_context
                                <https://scikit-learn.org/stable/modules/generated/
                                sklearn.config_context.html>`_, If True, validation for finiteness
                                will be skipped, saving time, but leading to potential crashes.
                                If False, validation for finiteness will be performed, avoiding error.
    @param      node_time       measure time execution for every node in the graph
    @param      versions        add columns with versions of used packages,
                                :epkg:`numpy`, :epkg:`scikit-learn`, :epkg:`onnx`,
                                :epkg:`onnxruntime`, :epkg:`sklearn-onnx`
    @param      extended_list   also check models this module implements a converter for
    @param      time_kwargs     to define a more precise way to measure a model
    @param      n_features      modifies the shorts datasets used to train the models
                                to use exactly this number of features, it can also
                                be a list to test multiple datasets
    @param      skip_long_test  skips tests for high values of N if they seem too long
    @param      fail_bad_results fails if the results are aligned with :epkg:`scikit-learn`
    @param      time_kwargs_fact see :func:`_multiply_time_kwargs <mlprodict.onnxrt.validate.validate_helper._multiply_time_kwargs>`
    @param      time_limit      to skip the rest of the test after this limit (in second)
    @param      n_jobs          *n_jobs* is set to the number of CPU by default unless this
                                value is changed
    @param      fLOG            logging function
    @return                     list of dictionaries

    The function is available through command line
    :ref:`validate_runtime <l-cmd-validate_runtime>`.
    The default for *time_kwargs* is the following:

    .. runpython::
        :showcode:
        :warningout: DeprecationWarning

        from mlprodict.onnxrt.validate.validate_helper import default_time_kwargs
        import pprint
        pprint.pprint(default_time_kwargs())
    """
    register_converters()
    register_rewritten_operators()
    ops = _enumerate_validated_operator_opsets_ops(
        extended_list, models, skip_models)

    if verbose > 0:

        def iterate():
            for i, row in enumerate(ops):  # pragma: no cover
                fLOG("{}/{} - {}".format(i + 1, len(ops), row))
                yield row

        if verbose >= 11:
            verbose -= 10  # pragma: no cover
            loop = iterate()  # pragma: no cover
        else:
            try:
                from tqdm import trange

                def iterate_tqdm():
                    with trange(len(ops)) as t:
                        for i in t:
                            row = ops[i]
                            disp = row['name'] + " " * (28 - len(row['name']))
                            t.set_description("%s" % disp)
                            yield row

                loop = iterate_tqdm()

            except ImportError:  # pragma: no cover
                loop = iterate()
    else:
        loop = ops

    if versions:
        add_versions = _enumerate_validated_operator_opsets_version(runtime)
    else:
        add_versions = {}

    current_opset = get_opset_number_from_onnx()
    if opset_min == -1:
        opset_min = get_opset_number_from_onnx()
    if opset_max == -1:
        opset_max = get_opset_number_from_onnx()
    if verbose > 0 and fLOG is not None:
        fLOG("[enumerate_validated_operator_opsets] opset in [{}, {}].".format(
            opset_min, opset_max))
    for row in loop:

        model = row['cl']
        if verbose > 1:
            fLOG("[enumerate_validated_operator_opsets] - model='{}'".format(model))

        for obs in enumerate_compatible_opset(
                model, opset_min=opset_min, opset_max=opset_max,
                check_runtime=check_runtime, runtime=runtime,
                debug=debug, dump_folder=dump_folder,
                store_models=store_models, benchmark=benchmark,
                fLOG=fLOG, filter_exp=filter_exp,
                assume_finite=assume_finite, node_time=node_time,
                verbose=verbose, extended_list=extended_list,
                time_kwargs=time_kwargs, dump_all=dump_all,
                n_features=n_features, skip_long_test=skip_long_test,
                filter_scenario=filter_scenario,
                time_kwargs_fact=time_kwargs_fact,
                time_limit=time_limit, n_jobs=n_jobs):

            for mandkey in ('inst', 'method_name', 'problem',
                            'scenario'):
                if '_0problem_exc' in obs:
                    continue
                if mandkey not in obs:
                    raise ValueError("Missing key '{}' in\n{}".format(
                        mandkey, pprint.pformat(obs)))  # pragma: no cover
            if verbose > 1:
                fLOG('[enumerate_validated_operator_opsets] - OBS')
                if verbose > 2:
                    fLOG("  ", obs)
                else:
                    obs_log = {k: v for k,
                               v in obs.items() if 'lambda-' not in k}
                    fLOG("  ", obs_log)
            elif verbose > 0 and "_0problem_exc" in obs:
                fLOG("  ???", obs)  # pragma: no cover

            diff = obs.get('max_rel_diff_batch', None)
            batch = 'max_rel_diff_batch' in obs and diff is not None
            op1 = obs.get('domain_opset_', '')
            op2 = obs.get('domain_opset_ai.onnx.ml', '')
            op = '{}/{}'.format(op1, op2)

            obs['available'] = "?"
            if diff is not None:
                if diff < 1e-5:
                    obs['available'] = 'OK'
                elif diff < 0.0001:
                    obs['available'] = 'e<0.0001'
                elif diff < 0.001:
                    obs['available'] = 'e<0.001'
                elif diff < 0.01:
                    obs['available'] = 'e<0.01'  # pragma: no cover
                elif diff < 0.1:
                    obs['available'] = 'e<0.1'
                else:
                    obs['available'] = "ERROR->=%1.1f" % diff
                obs['available'] += '-' + op
                if not batch:
                    obs['available'] += "-NOBATCH"  # pragma: no cover
                if fail_bad_results and 'e<' in obs['available']:
                    raise RuntimeBadResultsError(
                        "Wrong results '{}'.".format(obs['available']), obs)  # pragma: no cover

            excs = []
            for k, v in sorted(obs.items()):
                if k.endswith('_exc'):
                    excs.append((k, v))
                    break
            if 'opset' not in obs:
                # It fails before the conversion happens.
                obs['opset'] = current_opset
            if obs['opset'] == current_opset and len(excs) > 0:
                k, v = excs[0]
                obs['available'] = 'ERROR-%s' % k
                obs['available-ERROR'] = v

            if 'bench-skl' in obs:
                b1 = obs['bench-skl']
                if 'bench-batch' in obs:
                    b2 = obs['bench-batch']
                else:
                    b2 = None
                if b1 is not None and b2 is not None:
                    for k in b1:
                        if k in b2 and b2[k] is not None and b1[k] is not None:
                            key = 'time-ratio-N=%d' % k
                            obs[key] = b2[k]['average'] / b1[k]['average']
                            key = 'time-ratio-N=%d-min' % k
                            obs[key] = b2[k]['min_exec'] / b1[k]['max_exec']
                            key = 'time-ratio-N=%d-max' % k
                            obs[key] = b2[k]['max_exec'] / b1[k]['min_exec']

            obs.update(row)
            obs.update(add_versions)
            yield obs.copy()



"""
@file
@brief Inspired from skl2onnx, handles two backends.
"""
import numpy
import onnx
import pandas
from ...tools.ort_wrapper import OrtInvalidArgument
from .utils_backend_common import (
    load_data_and_model, extract_options,
    ExpectedAssertionError, OnnxBackendAssertionError,
    OnnxRuntimeMissingNewOnnxOperatorException,
    _compare_expected, _create_column)


def compare_runtime_session(  # pylint: disable=R0912
        cls_session, test, decimal=5, options=None,
        verbose=False, context=None, comparable_outputs=None,
        intermediate_steps=False, classes=None,
        disable_optimisation=False):
    """
    The function compares the expected output (computed with
    the model before being converted to ONNX) and the ONNX output
    produced with module :epkg:`onnxruntime` or :epkg:`mlprodict`.

    :param cls_session: inference session instance (like @see cl OnnxInference)
    :param test: dictionary with the following keys:
        - *onnx*: onnx model (filename or object)
        - *expected*: expected output (filename pkl or object)
        - *data*: input data (filename pkl or object)
    :param decimal: precision of the comparison
    :param options: comparison options
    :param context: specifies custom operators
    :param verbose: in case of error, the function may print
        more information on the standard output
    :param comparable_outputs: compare only these outputs
    :param intermediate_steps: displays intermediate steps
        in case of an error
    :param classes: classes names (if option 'nocl' is used)
    :param disable_optimisation: disable optimisation the runtime may do
    :return: tuple (outut, lambda function to run the predictions)

    The function does not return anything but raises an error
    if the comparison failed.
    """
    lambda_onnx = None
    if context is None:
        context = {}
    load = load_data_and_model(test, **context)
    if verbose:  # pragma no cover
        print("[compare_runtime] test '{}' loaded".format(test['onnx']))

    onx = test['onnx']

    if options is None:
        if isinstance(onx, str):
            options = extract_options(onx)
        else:
            options = {}
    elif options is None:
        options = {}
    elif not isinstance(options, dict):
        raise TypeError(  # pragma no cover
            "options must be a dictionary.")

    if verbose:  # pragma no cover
        print("[compare_runtime] InferenceSession('{}')".format(onx))

    runtime_options = dict(disable_optimisation=disable_optimisation)
    try:
        sess = cls_session(onx, runtime_options=runtime_options)
    except TypeError as e:  # pragma: no cover
        raise TypeError(  # pylint: disable=W0707
            "Wrong signature for '{}'.".format(cls_session.__name__))
    except ExpectedAssertionError as expe:  # pragma no cover
        raise expe
    except Exception as e:  # pylint: disable=W0703
        if "CannotLoad" in options:  # pragma no cover
            raise ExpectedAssertionError(  # pylint: disable=W0707
                "Unable to load onnx '{0}' due to\n{1}".format(onx, e))
        else:  # pragma no cover
            if verbose:  # pragma no cover
                model = onnx.load(onx)
                smodel = "\nJSON ONNX\n" + str(model)
            else:
                smodel = ""
            if ("NOT_IMPLEMENTED : Could not find an implementation "
                    "for the node" in str(e)):
                # onnxruntime does not implement a specific node yet.
                raise OnnxRuntimeMissingNewOnnxOperatorException(  # pylint: disable=W0707
                    "{3} does not implement a new operator "
                    "'{0}'\n{1}\nONNX\n{2}".format(
                        onx, e, smodel, cls_session))
            if "NOT_IMPLEMENTED : Failed to find kernel" in str(e):
                # onnxruntime does not implement a specific node yet
                # in the kernel included in onnxruntime.
                raise OnnxBackendAssertionError(  # pylint: disable=W0707
                    "{3} misses a kernel for operator "
                    "'{0}'\n{1}\nONNX\n{2}".format(
                        onx, e, smodel, cls_session))
            raise OnnxBackendAssertionError(  # pylint: disable=W0707
                "Unable to load onnx '{0}'\nONNX\n{1}\n{2}".format(
                    onx, smodel, e))

    input = load["data"]
    DF = options.pop('DF', False)
    if DF:
        inputs = {c: input[c].values for c in input.columns}
        for k in inputs:
            if inputs[k].dtype == numpy.float64:
                inputs[k] = inputs[k].astype(numpy.float32)
            inputs[k] = inputs[k].reshape((inputs[k].shape[0], 1))
    else:
        if isinstance(input, dict):
            inputs = input
        elif isinstance(input, (list, numpy.ndarray, pandas.DataFrame)):
            inp = sess.get_inputs()
            outs = sess.get_outputs()
            if len(outs) == 0:
                raise OnnxBackendAssertionError(  # pragma: no cover
                    "Wrong number of outputs, onnx='{2}'".format(onx))
            if len(inp) == len(input):
                inputs = {i.name: v for i, v in zip(inp, input)}
            elif len(inp) == 1:
                inputs = {inp[0].name: input}
            elif isinstance(input, numpy.ndarray):
                shape = sum(i.shape[1] if len(i.shape) == 2 else i.shape[0]
                            for i in inp)
                if shape == input.shape[1]:
                    inputs = {n.name: input[:, i] for i, n in enumerate(inp)}
                else:
                    raise OnnxBackendAssertionError(  # pragma: no cover
                        "Wrong number of inputs onnx {0} != "
                        "original shape {1}, onnx='{2}'"
                        .format(len(inp), input.shape, onx))
            elif isinstance(input, list):
                try:
                    array_input = numpy.array(input)
                except Exception:  # pragma no cover
                    raise OnnxBackendAssertionError(  # pylint: disable=W0707
                        "Wrong number of inputs onnx {0} != "
                        "original {1}, onnx='{2}'"
                        .format(len(inp), len(input), onx))
                shape = sum(i.shape[1] for i in inp)
                if shape == array_input.shape[1]:
                    inputs = {}
                    c = 0
                    for i, n in enumerate(inp):
                        d = c + n.shape[1]
                        inputs[n.name] = _create_column(
                            [row[c:d] for row in input], n.type)
                        c = d
                else:
                    raise OnnxBackendAssertionError(  # pragma no cover
                        "Wrong number of inputs onnx {0} != "
                        "original shape {1}, onnx='{2}'*"
                        .format(len(inp), array_input.shape, onx))
            elif isinstance(input, pandas.DataFrame):
                try:
                    array_input = numpy.array(input)
                except Exception:  # pragma no cover
                    raise OnnxBackendAssertionError(  # pylint: disable=W0707
                        "Wrong number of inputs onnx {0} != "
                        "original {1}, onnx='{2}'"
                        .format(len(inp), len(input), onx))
                shape = sum(i.shape[1] for i in inp)
                if shape == array_input.shape[1]:
                    inputs = {}
                    c = 0
                    for i, n in enumerate(inp):
                        d = c + n.shape[1]
                        inputs[n.name] = _create_column(
                            input.iloc[:, c:d], n.type)
                        c = d
                else:
                    raise OnnxBackendAssertionError(  # pragma no cover
                        "Wrong number of inputs onnx {0}={1} columns != "
                        "original shape {2}, onnx='{3}'*"
                        .format(len(inp), shape, array_input.shape, onx))
            else:
                raise OnnxBackendAssertionError(  # pragma no cover
                    "Wrong type of inputs onnx {0}, onnx='{1}'".format(
                        type(input), onx))
        else:
            raise OnnxBackendAssertionError(  # pragma no cover
                "Dict or list is expected, not {0}".format(type(input)))

        for k in inputs:
            if isinstance(inputs[k], list):
                inputs[k] = numpy.array(inputs[k])

    options.pop('SklCol', False)  # unused here but in dump_data_and_model

    if verbose:  # pragma no cover
        print("[compare_runtime] type(inputs)={} len={} names={}".format(
            type(input), len(inputs), list(sorted(inputs))))
    if verbose:  # pragma no cover
        if intermediate_steps:
            run_options = {'verbose': 3, 'fLOG': print}
        else:
            run_options = {'verbose': 2, 'fLOG': print}
    else:
        run_options = {}
    try:
        try:
            output = sess.run(None, inputs, **run_options)
        except TypeError:  # pragma no cover
            output = sess.run(None, inputs)
        lambda_onnx = lambda: sess.run(None, inputs)  # noqa
        if verbose:  # pragma no cover
            import pprint
            pprint.pprint(output)
    except ExpectedAssertionError as expe:  # pragma no cover
        raise expe
    except (RuntimeError, OrtInvalidArgument) as e:  # pragma no cover
        if intermediate_steps:
            sess.run(None, inputs, verbose=3, fLOG=print)
        if "-Fail" in onx:
            raise ExpectedAssertionError(  # pylint: disable=W0707
                "{1} cannot compute the prediction for '{0}'".
                format(onx, cls_session))
        else:
            if verbose:  # pragma no cover
                model = onnx.load(onx)
                smodel = "\nJSON ONNX\n" + str(model)
            else:
                smodel = ""
            import pprint
            raise OnnxBackendAssertionError(  # pylint: disable=W0707
                "{4} cannot compute the predictions"
                " for '{0}' due to {1}{2}\n{3}"
                .format(onx, e, smodel, pprint.pformat(inputs),
                        cls_session))
    except Exception as e:  # pragma no cover
        raise OnnxBackendAssertionError(  # pylint: disable=W0707
            "Unable to run onnx '{0}' due to {1}".format(onx, e))
    if verbose:  # pragma no cover
        print("[compare_runtime] done type={}".format(type(output)))

    output0 = output.copy()

    if comparable_outputs:
        cmp_exp = [load["expected"][o] for o in comparable_outputs]
        cmp_out = [output[o] for o in comparable_outputs]
    else:
        cmp_exp = load["expected"]
        cmp_out = output

    try:
        _compare_expected(cmp_exp, cmp_out, sess, onx,
                          decimal=decimal, verbose=verbose,
                          classes=classes, **options)
    except ExpectedAssertionError as expe:  # pragma no cover
        raise expe
    except Exception as e:  # pragma no cover
        if verbose:  # pragma no cover
            model = onnx.load(onx)
            smodel = "\nJSON ONNX\n" + str(model)
        else:
            smodel = ""
        raise OnnxBackendAssertionError(  # pylint: disable=W0707
            "Model '{}' has discrepencies with cls='{}'.\n{}: {}{}".format(
                onx, sess.__class__.__name__, type(e), e, smodel))

    return output0, lambda_onnx



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import unicodedata
import locale
import warnings
import numpy
from ._op import OpRunUnary, RuntimeTypeError


class StringNormalizer(OpRunUnary):
    """
    The operator is not really threadsafe as python cannot
    play with two locales at the same time. stop words
    should not be implemented here as the tokenization
    usually happens after this steps.
    """

    atts = {'case_change_action': b'NONE',  # LOWER UPPER NONE
            'is_case_sensitive': 1,
            'locale': b'',
            'stopwords': []}

    def __init__(self, onnx_node, desc=None, **options):
        OpRunUnary.__init__(self, onnx_node, desc=desc,
                            expected_attributes=StringNormalizer.atts,
                            **options)
        self.slocale = self.locale.decode('ascii')
        self.stops = set(self.stopwords)

    def _run(self, x):  # pylint: disable=W0221
        """
        Normalizes strings.
        """
        res = numpy.empty(x.shape, dtype=x.dtype)
        if len(x.shape) == 2:
            for i in range(0, x.shape[1]):
                self._run_column(x[:, i], res[:, i])
        elif len(x.shape) == 1:
            self._run_column(x, res)
        else:
            raise RuntimeTypeError(  # pragma: no cover
                "x must be a matrix or a vector.")
        return (res, )

    def _run_column(self, cin, cout):
        """
        Normalizes string in a columns.
        """
        if locale.getlocale() != self.slocale:
            try:
                locale.setlocale(locale.LC_ALL, self.slocale)
            except locale.Error as e:
                warnings.warn(
                    "Unknown local setting '{}' (current: '{}') - {}."
                    "".format(self.slocale, locale.getlocale(), e))
        stops = set(_.decode() for _ in self.stops)
        cout[:] = cin[:]

        for i in range(0, cin.shape[0]):
            cout[i] = self.strip_accents_unicode(cout[i])

        if self.is_case_sensitive and len(stops) > 0:
            for i in range(0, cin.shape[0]):
                cout[i] = self._remove_stopwords(cout[i], stops)

        if self.case_change_action == b'LOWER':
            for i in range(0, cin.shape[0]):
                cout[i] = cout[i].lower()
        elif self.case_change_action == b'UPPER':
            for i in range(0, cin.shape[0]):
                cout[i] = cout[i].upper()
        elif self.case_change_action != b'NONE':
            raise RuntimeError(
                "Unknown option for case_change_action: {}.".format(
                    self.case_change_action))

        if not self.is_case_sensitive and len(stops) > 0:
            for i in range(0, cin.shape[0]):
                cout[i] = self._remove_stopwords(cout[i], stops)

        return cout

    def _remove_stopwords(self, text, stops):
        spl = text.split(' ')
        return ' '.join(filter(lambda s: s not in stops, spl))

    def strip_accents_unicode(self, s):
        """
        Transforms accentuated unicode symbols into their simple counterpart.
        Source: `sklearn/feature_extraction/text.py
        <https://github.com/scikit-learn/scikit-learn/blob/master/sklearn/
        feature_extraction/text.py#L115>`_.

        :param s: string
            The string to strip
        :return: the cleaned string
        """
        try:
            # If `s` is ASCII-compatible, then it does not contain any accented
            # characters and we can avoid an expensive list comprehension
            s.encode("ASCII", errors="strict")
            return s
        except UnicodeEncodeError:
            normalized = unicodedata.normalize('NFKD', s)
            s = ''.join(
                [c for c in normalized if not unicodedata.combining(c)])
            return s

    def _infer_shapes(self, x):  # pylint: disable=E0202,W0221
        return (x, )



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRunUnaryNum


class Neg(OpRunUnaryNum):

    def __init__(self, onnx_node, desc=None, **options):
        OpRunUnaryNum.__init__(self, onnx_node, desc=desc,
                               expected_attributes=None,
                               **options)

    def _run(self, data):  # pylint: disable=W0221
        if self.inplaces.get(0, False):
            numpy.negative(data, out=data)
        else:
            data = numpy.negative(data)
        return (data, )

    def to_python(self, inputs):
        return ("import numpy",
                "return -%s" % inputs[0])



"""
@file
@brief Inspired from skl2onnx, handles two backends.
"""
from pyquickhelper.pycode import is_travis_or_appveyor
from .utils_backend_common_compare import compare_runtime_session
from ...tools.ort_wrapper import (
    InferenceSession, GraphOptimizationLevel, SessionOptions)


def _capture_output(fct, kind):
    if is_travis_or_appveyor():
        return fct(), None, None  # pragma: no cover
    try:
        from cpyquickhelper.io import capture_output
    except ImportError:
        # cpyquickhelper not available
        return fct(), None, None  # pragma: no cover
    return capture_output(fct, kind)  # pragma: no cover


class InferenceSession2:
    """
    Overwrites class *InferenceSession* to capture
    the standard output and error.
    """

    def __init__(self, *args, **kwargs):
        "Overwrites the constructor."
        runtime_options = kwargs.pop('runtime_options', {})
        disable_optimisation = runtime_options.pop(
            'disable_optimisation', False)
        if disable_optimisation:
            if 'sess_options' in kwargs:
                raise RuntimeError(  # pragma: no cover
                    "Incompatible options, 'disable_options' and 'sess_options' cannot "
                    "be sepcified at the same time.")
            kwargs['sess_options'] = SessionOptions()
            kwargs['sess_options'].graph_optimization_level = (
                GraphOptimizationLevel.ORT_DISABLE_ALL)
        self.sess, self.outi, self.erri = _capture_output(
            lambda: InferenceSession(*args, **kwargs), 'c')

    def run(self, *args, **kwargs):
        "Overwrites method *run*."
        res, self.outr, self.errr = _capture_output(
            lambda: self.sess.run(*args, **kwargs), 'c')
        return res

    def get_inputs(self, *args, **kwargs):
        "Overwrites method *get_inputs*."
        return self.sess.get_inputs(*args, **kwargs)

    def get_outputs(self, *args, **kwargs):
        "Overwrites method *get_outputs*."
        return self.sess.get_outputs(*args, **kwargs)


def compare_runtime(test, decimal=5, options=None,
                    verbose=False, context=None, comparable_outputs=None,
                    intermediate_steps=False, classes=None,
                    disable_optimisation=False):
    """
    The function compares the expected output (computed with
    the model before being converted to ONNX) and the ONNX output
    produced with module :epkg:`onnxruntime` or :epkg:`mlprodict`.

    :param test: dictionary with the following keys:
        - *onnx*: onnx model (filename or object)
        - *expected*: expected output (filename pkl or object)
        - *data*: input data (filename pkl or object)
    :param decimal: precision of the comparison
    :param options: comparison options
    :param context: specifies custom operators
    :param verbose: in case of error, the function may print
        more information on the standard output
    :param comparable_outputs: compare only these outputs
    :param intermediate_steps: displays intermediate steps
        in case of an error
    :param classes: classes names (if option 'nocl' is used)
    :param disable_optimisation: disable optimisation onnxruntime
        could do
    :return: tuple (outut, lambda function to run the predictions)

    The function does not return anything but raises an error
    if the comparison failed.
    """
    return compare_runtime_session(
        InferenceSession2, test, decimal=decimal, options=options,
        verbose=verbose, context=context,
        comparable_outputs=comparable_outputs,
        intermediate_steps=intermediate_steps,
        classes=classes, disable_optimisation=disable_optimisation)



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRunUnaryNum, RuntimeTypeError


class Imputer(OpRunUnaryNum):

    atts = {'imputed_value_floats': numpy.empty(0, dtype=numpy.float32),
            'imputed_value_int64s': numpy.empty(0, dtype=numpy.int64),
            'replaced_value_float': 0.,
            'replaced_value_int64': 0}

    def __init__(self, onnx_node, desc=None, **options):
        OpRunUnaryNum.__init__(self, onnx_node, desc=desc,
                               expected_attributes=Imputer.atts,
                               **options)
        if len(self.imputed_value_floats) > 0:
            self.values = self.imputed_value_floats
            self.replace = self.replaced_value_float
        elif len(self.imputed_value_int64s) > 0:
            self.values = self.imputed_value_int64s
            self.replace = self.replaced_value_int64
        else:
            raise ValueError("Missing are not defined.")  # pragma: no cover

    def _run(self, x):  # pylint: disable=W0221
        if len(x.shape) != 2:
            raise RuntimeTypeError(
                "x must be a matrix but shape is {}".format(x.shape))
        if self.values.shape[0] not in (x.shape[1], 1):
            raise RuntimeTypeError(  # pragma: no cover
                "Dimension mismatch {} != {}".format(
                    self.values.shape[0], x.shape[1]))
        x = x.copy()
        if numpy.isnan(self.replace):
            for i in range(0, x.shape[1]):
                val = self.values[min(i, self.values.shape[0] - 1)]
                x[numpy.isnan(x[:, i]), i] = val
        else:
            for i in range(0, x.shape[1]):
                val = self.values[min(i, self.values.shape[0] - 1)]
                x[x[:, i] == self.replace, i] = val

        return (x, )



# -*- encoding: utf-8 -*-
"""
@file
@brief Shortcut to *ops_whole*.
"""
import json
from io import BytesIO
import onnx
from ...tools.ort_wrapper import (
    InferenceSession, SessionOptions, RunOptions,
    GraphOptimizationLevel, OrtFail,
    OrtInvalidGraph, OrtInvalidArgument,
    OrtNotImplemented, OrtRuntimeException)
from ...tools.asv_options_helper import display_onnx


class OnnxWholeSession:
    """
    Runs the prediction for a single :epkg:`ONNX`,
    it lets the runtime handle the graph logic as well.
    """

    def __init__(self, onnx_data, runtime, runtime_options=None):
        """
        @param      onnx_data       :epkg:`ONNX` model or data
        @param      runtime         runtime to be used,
                                    mostly :epkg:`onnxruntime`
        @param      runtime_options runtime options
        """
        if runtime != 'onnxruntime1':
            raise NotImplementedError(  # pragma: no cover
                "runtime '{}' is not implemented.".format(runtime))
        if hasattr(onnx_data, 'SerializeToString'):
            onnx_data = onnx_data.SerializeToString()
        session_options = (
            None if runtime_options is None
            else runtime_options.get('session_options', None))
        self.runtime = runtime
        sess_options = session_options or SessionOptions()
        self.run_options = RunOptions()

        if session_options is None:
            try:
                sess_options.sessions_log_verbosity_level = 0
            except AttributeError:  # pragma: no cover
                # onnxruntime not recent enough.
                pass
            try:
                self.run_options.run_log_verbosity_level = 0
            except AttributeError:  # pragma: no cover
                # onnxruntime not recent enough.
                pass
            if runtime_options is not None:
                if runtime_options.get('disable_optimisation', False):
                    sess_options.graph_optimization_level = (  # pragma: no cover
                        GraphOptimizationLevel.ORT_ENABLE_ALL)
                if runtime_options.get('enable_profiling', True):
                    sess_options.enable_profiling = True
        elif 'enable_profiling' in runtime_options:
            raise RuntimeError(  # pragma: no cover
                "session_options and enable_profiling cannot be defined at the "
                "same time.")
        elif 'disable_optimisation' in runtime_options:
            raise RuntimeError(  # pragma: no cover
                "session_options and disable_optimisation cannot be defined at the "
                "same time.")
        try:
            self.sess = InferenceSession(onnx_data, sess_options=sess_options)
        except (OrtFail, OrtNotImplemented, OrtInvalidGraph,
                OrtInvalidArgument, OrtRuntimeException, RuntimeError) as e:
            raise RuntimeError(
                "Unable to create InferenceSession due to '{}'\n{}.".format(
                    e, display_onnx(onnx.load(BytesIO(onnx_data))))) from e

    def run(self, inputs):
        """
        Computes the predictions.

        @param      inputs      dictionary *{variable, value}*
        @return                 list of outputs
        """
        return self.sess.run(None, inputs, self.run_options)

    @staticmethod
    def process_profiling(js):
        """
        Flattens json returned by onnxruntime profiling.

        :param js: json
        :return: list of dictionaries
        """
        rows = []
        for row in js:
            if 'args' in row and isinstance(row['args'], dict):
                for k, v in row['args'].items():
                    row['args_%s' % k] = v
                del row['args']
            rows.append(row)
        return rows

    def get_profiling(self):
        """
        Returns the profiling informations.
        """
        prof = self.sess.end_profiling()
        with open(prof, 'r') as f:
            content = f.read()
        js = json.loads(content)
        return OnnxWholeSession.process_profiling(js)



"""
@file
@brief numpy redundant functions.
"""
import numpy
from scipy.sparse.coo import coo_matrix


def numpy_dot_inplace(inplaces, a, b):
    """
    Implements a dot product, deals with inplace information.
    See :epkg:`numpy:dot`.
    """
    if inplaces.get(0, False) and hasattr(a, 'flags'):
        return _numpy_dot_inplace_left(a, b)
    if inplaces.get(1, False) and hasattr(b, 'flags'):
        return _numpy_dot_inplace_right(a, b)
    return numpy.dot(a, b)


def _numpy_dot_inplace_left(a, b):
    "Subpart of @see fn numpy_dot_inplace."
    if a.flags['F_CONTIGUOUS']:
        if len(b.shape) == len(a.shape) == 2 and b.shape[1] <= a.shape[1]:
            try:
                numpy.dot(a, b, out=a[:, :b.shape[1]])
                return a[:, :b.shape[1]]
            except ValueError:
                return numpy.dot(a, b)
        if len(b.shape) == 1:
            try:
                numpy.dot(a, b.reshape(b.shape[0], 1), out=a[:, :1])
                return a[:, :1].reshape(a.shape[0])
            except ValueError:  # pragma no cover
                return numpy.dot(a, b)
    return numpy.dot(a, b)


def _numpy_dot_inplace_right(a, b):
    "Subpart of @see fn numpy_dot_inplace."
    if b.flags['C_CONTIGUOUS']:
        if len(b.shape) == len(a.shape) == 2 and a.shape[0] <= b.shape[0]:
            try:
                numpy.dot(a, b, out=b[:a.shape[0], :])
                return b[:a.shape[0], :]
            except ValueError:  # pragma no cover
                return numpy.dot(a, b)
        if len(a.shape) == 1:
            try:
                numpy.dot(a, b, out=b[:1, :])
                return b[:1, :]
            except ValueError:  # pragma no cover
                return numpy.dot(a, b)
    return numpy.dot(a, b)


def numpy_matmul_inplace(inplaces, a, b):
    """
    Implements a matmul product, deals with inplace information.
    See :epkg:`numpy:matmul`.
    Inplace computation does not work well as modifying one of the
    container modifies the results. This part still needs to be
    improves.
    """
    try:
        if isinstance(a, coo_matrix) or isinstance(b, coo_matrix):
            return numpy.dot(a, b)
        if len(a.shape) <= 2 and len(b.shape) <= 2:
            return numpy_dot_inplace(inplaces, a, b)
        return numpy.matmul(a, b)
    except ValueError as e:
        raise ValueError(
            "Unable to multiply shapes %r, %r." % (a.shape, b.shape)) from e



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from scipy.special import expit  # pylint: disable=E0611
from ._op import OpRunClassifierProb
from ._op_classifier_string import _ClassifierCommon
from ._op_numpy_helper import numpy_dot_inplace


class LinearClassifier(OpRunClassifierProb, _ClassifierCommon):

    atts = {'classlabels_ints': [], 'classlabels_strings': [],
            'coefficients': None, 'intercepts': None,
            'multi_class': 0, 'post_transform': b'NONE'}

    def __init__(self, onnx_node, desc=None, **options):
        OpRunClassifierProb.__init__(self, onnx_node, desc=desc,
                                     expected_attributes=LinearClassifier.atts,
                                     **options)
        self._post_process_label_attributes()
        if not isinstance(self.coefficients, numpy.ndarray):
            raise TypeError("coefficient must be an array not {}.".format(
                type(self.coefficients)))
        if len(getattr(self, "classlabels_ints", [])) == 0 and \
                len(getattr(self, 'classlabels_strings', [])) == 0:
            raise ValueError(
                "Fields classlabels_ints or classlabels_strings must be specified.")
        self.nb_class = max(len(getattr(self, 'classlabels_ints', [])),
                            len(getattr(self, 'classlabels_strings', [])))
        if len(self.coefficients.shape) != 1:
            raise ValueError("coefficient must be an array but has shape {}\n{}.".format(
                self.coefficients.shape, desc))
        n = self.coefficients.shape[0] // self.nb_class
        self.coefficients = self.coefficients.reshape(self.nb_class, n).T

    def _run(self, x):  # pylint: disable=W0221
        scores = numpy_dot_inplace(self.inplaces, x, self.coefficients)
        if self.intercepts is not None:
            scores += self.intercepts

        if self.post_transform == b'NONE':
            pass
        elif self.post_transform == b'LOGISTIC':
            expit(scores, out=scores)
        elif self.post_transform == b'SOFTMAX':
            numpy.subtract(scores, scores.max(axis=1)[
                           :, numpy.newaxis], out=scores)
            numpy.exp(scores, out=scores)
            numpy.divide(scores, scores.sum(axis=1)[
                         :, numpy.newaxis], out=scores)
        else:
            raise NotImplementedError("Unknown post_transform: '{}'.".format(
                self.post_transform))

        if self.nb_class == 1:
            label = numpy.zeros((scores.shape[0],), dtype=x.dtype)
            label[scores > 0] = 1
        else:
            label = numpy.argmax(scores, axis=1)
        return self._post_process_predicted_label(label, scores)



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRunReduceNumpy


class ReduceSumSquare(OpRunReduceNumpy):

    atts = {'axes': [], 'keepdims': 1}

    def __init__(self, onnx_node, desc=None, **options):
        OpRunReduceNumpy.__init__(self, onnx_node, desc=desc,
                                  expected_attributes=ReduceSumSquare.atts,
                                  **options)

    def _run(self, data):  # pylint: disable=W0221
        return (numpy.sum(numpy.square(data), axis=self.axes,
                          keepdims=self.keepdims), )



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import re
import numpy
from ._op import OpRunUnary, RuntimeTypeError
from ._new_ops import OperatorSchema
from ..shape_object import ShapeObject


class Tokenizer(OpRunUnary):
    """
    See :epkg:`Tokenizer`.
    """

    atts = {'mark': 0,
            'mincharnum': 1,
            'pad_value': b'#',
            'separators': [],
            'tokenexp': b'[a-zA-Z0-9_]+',
            'tokenexpsplit': 0,
            'stopwords': []}

    def __init__(self, onnx_node, desc=None, **options):
        OpRunUnary.__init__(self, onnx_node, desc=desc,
                            expected_attributes=Tokenizer.atts,
                            **options)
        self.char_tokenization_ = (
            self.tokenexp == b'.' or list(self.separators) == [b''])
        self.stops_ = set(_.decode() for _ in self.stopwords)
        try:
            self.str_separators_ = set(_.decode('utf-8')
                                       for _ in self.separators)
        except AttributeError as e:  # pragma: no cover
            raise RuntimeTypeError(
                "Unable to interpret separators {}.".format(self.separators)) from e
        if self.tokenexp not in (None, b''):
            self.tokenexp_ = re.compile(self.tokenexp.decode('utf-8'))

    def _find_custom_operator_schema(self, op_name):
        if op_name == "Tokenizer":
            return TokenizerSchema()
        raise RuntimeError(  # pragma: no cover
            "Unable to find a schema for operator '{}'.".format(op_name))

    def _run(self, text):  # pylint: disable=W0221
        if self.char_tokenization_:
            return self._run_char_tokenization(text, self.stops_)
        if self.str_separators_ is not None and len(self.str_separators_) > 0:
            return self._run_sep_tokenization(
                text, self.stops_, self.str_separators_)
        if self.tokenexp not in (None, ''):
            return self._run_regex_tokenization(
                text, self.stops_, self.tokenexp_)
        raise RuntimeError(  # pragma: no cover
            "Unable to guess which tokenization to use, sep={}, "
            "tokenexp='{}'.".format(self.separators, self.tokenexp))

    def _run_tokenization(self, text, stops, split):
        """
        Tokenizes a char level.
        """
        max_len = max(map(len, text.flatten()))
        if self.mark:
            max_len += 2
            begin = 1
        else:
            begin = 0
        shape = text.shape + (max_len, )
        max_pos = 0
        res = numpy.empty(shape, dtype=text.dtype)
        if len(text.shape) == 1:
            res[:] = self.pad_value
            for i in range(text.shape[0]):
                pos = begin
                for c in split(text[i]):
                    if c not in stops:
                        res[i, pos] = c
                        pos += 1
                if self.mark:
                    res[i, 0] = self.pad_value
                    max_pos = max(pos + 1, max_pos)
                else:
                    max_pos = max(pos, max_pos)
            res = res[:, :max_pos]
        elif len(text.shape) == 2:
            res[:, :] = self.pad_value
            for i in range(text.shape[0]):
                for ii in range(text.shape[1]):
                    pos = begin
                    for c in split(text[i, ii]):
                        if c not in stops:
                            res[i, ii, pos] = c
                            pos += 1
                    if self.mark:
                        res[i, ii, 0] = self.pad_value
                        max_pos = max(pos + 1, max_pos)
                    else:
                        max_pos = max(pos, max_pos)
            res = res[:, :, :max_pos]
        else:
            raise RuntimeError(  # pragma: no cover
                "Only vector or matrices are supported not shape {}.".format(text.shape))
        return (res, )

    def _run_char_tokenization(self, text, stops):
        """
        Tokenizes y charaters.
        """
        def split(t):
            for c in t:
                yield c
        return self._run_tokenization(text, stops, split)

    def _run_sep_tokenization(self, text, stops, separators):
        """
        Tokenizes using separators.
        The function should use a trie to find text.
        """
        def split(t):
            begin = 0
            pos = 0
            while pos < len(t):
                for sep in separators:
                    if (pos + len(sep) <= len(t) and
                            sep == t[pos: pos + len(sep)]):
                        word = t[begin: pos]
                        yield word
                        begin = pos + len(sep)
                        break
                pos += 1
            if begin < pos:
                word = t[begin: pos]
                yield word

        return self._run_tokenization(text, stops, split)

    def _run_regex_tokenization(self, text, stops, exp):
        """
        Tokenizes using separators.
        The function should use a trie to find text.
        """
        if self.tokenexpsplit:
            def split(t):
                return filter(lambda x: x, exp.split(t))
        else:
            def split(t):
                return filter(lambda x: x, exp.findall(t))
        return self._run_tokenization(text, stops, split)

    def _infer_shapes(self, x):  # pylint: disable=E0202,W0221
        if x.shape is None:
            return (x, )
        if len(x) == 1:
            return (ShapeObject((x[0], None), dtype=x.dtype,
                                name=self.__class__.__name__), )
        if len(x) == 2:
            return (ShapeObject((x[0], x[1], None), dtype=x.dtype,
                                name=self.__class__.__name__), )
        raise RuntimeTypeError(  # pragma: no cover
            "Only two dimension are allowed, got {}.".format(x))

    def _infer_types(self, x):  # pylint: disable=E0202,W0221
        return (x, )


class TokenizerSchema(OperatorSchema):
    """
    Defines a schema for operators added in this package
    such as @see cl TreeEnsembleClassifierDouble.
    """

    def __init__(self):
        OperatorSchema.__init__(self, 'Tokenizer')
        self.attributes = Tokenizer.atts



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from onnx.onnx_pb import TensorProto
from ._op import OpRun


class Cast(OpRun):

    atts = {'to': None}

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       expected_attributes=Cast.atts,
                       **options)
        # type help(TensorProto) to see all the possible values
        if self.to == TensorProto.FLOAT:  # pylint: disable=E1101
            self._dtype = numpy.float32
        elif self.to == TensorProto.DOUBLE:  # pylint: disable=E1101
            self._dtype = numpy.float64
        elif self.to == TensorProto.UINT8:  # pylint: disable=E1101
            self._dtype = numpy.uint8
        elif self.to == TensorProto.INT8:  # pylint: disable=E1101
            self._dtype = numpy.int8
        elif self.to == TensorProto.INT16:  # pylint: disable=E1101
            self._dtype = numpy.int16
        elif self.to == TensorProto.INT32:  # pylint: disable=E1101
            self._dtype = numpy.int32
        elif self.to == TensorProto.INT64:  # pylint: disable=E1101
            self._dtype = numpy.int64
        elif self.to == TensorProto.UINT16:  # pylint: disable=E1101
            self._dtype = numpy.uint16
        elif self.to == TensorProto.UINT32:  # pylint: disable=E1101
            self._dtype = numpy.uint32
        elif self.to == TensorProto.UINT64:  # pylint: disable=E1101
            self._dtype = numpy.uint64
        elif self.to == TensorProto.BOOL:  # pylint: disable=E1101
            self._dtype = numpy.bool_
        elif self.to == TensorProto.STRING:  # pylint: disable=E1101
            self._dtype = numpy.str_
        elif self.to == TensorProto.FLOAT16:  # pylint: disable=E1101
            self._dtype = numpy.float16
        elif self.to == TensorProto.COMPLEX64:  # pylint: disable=E1101
            self._dtype = numpy.complex64
        elif self.to == TensorProto.COMPLEX128:  # pylint: disable=E1101
            self._dtype = numpy.complex128
        else:
            raise ValueError(  # pragma: no cover
                "Unexpected value for to='{}'.".format(
                    self.to))  # pylint: disable=E1101
        self._cast = lambda x: x.astype(self._dtype)

    def _run(self, x):  # pylint: disable=W0221
        if self.inplaces.get(0, False):
            return self._run_inplace(x)
        return (self._cast(x), )

    def _run_inplace(self, x):
        if x.dtype == self._dtype:
            return (x, )
        return (self._cast(x), )

    def _infer_shapes(self, x):  # pylint: disable=W0221
        return (x.copy(dtype=self._dtype), )

    def _infer_types(self, x):  # pylint: disable=W0221
        return (self._dtype, )

    def _infer_sizes(self, *args, **kwargs):
        res = self.run(*args, **kwargs)
        return (dict(temp=0), ) + res



"""
@file
@brief Command line about validation of prediction runtime.
"""
from pandas import DataFrame


def benchmark_replay(folder, runtime='python', time_kwargs=None,
                     skip_long_test=True, time_kwargs_fact=None,
                     time_limit=4, out=None, verbose=1, fLOG=print):
    """
    The command rerun a benchmark if models were stored by
    command line `vaidate_runtime`.

    :param folder: where to find pickled files
    :param runtime: runtimes, comma separated list
    :param verbose: integer from 0 (None) to 2 (full verbose)
    :param out: output raw results into this file (excel format)
    :param time_kwargs: a dictionary which defines the number of rows and
        the parameter *number* and *repeat* when benchmarking a model,
        the value must follow :epkg:`json` format
    :param skip_long_test: skips tests for high values of N if
        they seem too long
    :param time_kwargs_fact: to multiply number and repeat in
        *time_kwargs* depending on the model
        (see :func:`_multiply_time_kwargs <mlprodict.onnxrt.validate.validate_helper._multiply_time_kwargs>`)
    :param time_limit: to stop benchmarking after this limit of time
    :param fLOG: logging function

    .. cmdref::
        :title: Replay a benchmark of stored converted models by validate_runtime
        :cmd: -m mlprodict benchmark_replay --help
        :lid: l-cmd-benchmark_replay

        The command rerun a benchmark if models were stored by
        command line `vaidate_runtime`.

        Example::

            python -m mlprodict benchmark_replay --folder dumped --out bench_results.xlsx

        Parameter ``--time_kwargs`` may be used to reduce or increase
        bencharmak precisions. The following value tells the function
        to run a benchmarks with datasets of 1 or 10 number, to repeat
        a given number of time *number* predictions in one row.
        The total time is divided by :math:`number \\times repeat``.
        Parameter ``--time_kwargs_fact`` may be used to increase these
        number for some specific models. ``'lin'`` multiplies
        by 10 number when the model is linear.

        ::

            -t "{\\"1\\":{\\"number\\":10,\\"repeat\\":10},\\"10\\":{\\"number\\":5,\\"repeat\\":5}}"
    """
    from ..onnxrt.validate.validate_benchmark_replay import enumerate_benchmark_replay  # pylint: disable=E0402

    rows = list(enumerate_benchmark_replay(
        folder=folder, runtime=runtime, time_kwargs=time_kwargs,
        skip_long_test=skip_long_test, time_kwargs_fact=time_kwargs_fact,
        time_limit=time_limit, verbose=verbose, fLOG=fLOG))
    if out is not None:
        df = DataFrame(rows)
        df.to_excel(out, index=False)
    return rows



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRunUnaryNum


def pycelu(x, alpha=1.):
    """
    Computes function ``celu(x)``.

    .. math::

        celu(x) = \\left \\{\\begin{array}{ll} x \\text{ if } x > 0 \\\\
        \\alpha ( e^{\\frac{x}{\\alpha}} - 1) \\, \\text{ otherwise }
        \\end{array} \\right.
    """
    if x > 0:
        return x
    return (numpy.exp(x / alpha) - 1) * alpha


def _vcelu1(x, alpha=1.):
    positive_input = numpy.maximum(0, x)
    negative_input = numpy.minimum(0, alpha * (
        numpy.exp(x / alpha) - 1))
    return positive_input + negative_input


class Celu(OpRunUnaryNum):

    atts = {'alpha': numpy.float32(1.0)}

    def __init__(self, onnx_node, desc=None, **options):
        OpRunUnaryNum.__init__(self, onnx_node, desc=desc,
                               expected_attributes=Celu.atts,
                               **options)
        self._vcelu2 = numpy.vectorize(
            lambda x: pycelu(x, self.alpha), otypes=[numpy.float])

    def _run(self, x):  # pylint: disable=W0221
        if self.inplaces.get(0, False):
            return self._run_inplace(x)
        return (_vcelu1(x, self.alpha), )

    def _run_inplace(self, x):
        return (self._vcelu2(x), )

    def to_python(self, inputs):
        return ('from mlprodict.onnxrt.ops_cpu.op_celu import _vcelu1',
                "return _vcelu1(X, alpha)")



"""
@file
@brief Summarizes results produces by function in *validate.py*.
"""
import decimal
import json
import numpy
import pandas
from sklearn import __all__ as sklearn__all__, __version__ as sklearn_version
from ... import __version__ as ort_version


def _clean_values_optim(val):
    if not isinstance(val, str):
        return val
    if '/' in val:
        spl = val.split('/')
        return "/".join(_clean_values_optim(v) for v in spl)
    if "'>=" in val:
        val = val.split("'>=")
        if len(val) == 2:
            val = val[-1]
    rep = {
        "{'optim': 'cdist'}": "cdist"
    }
    for k, v in rep.items():
        val = val.replace(k, v)
    return val


def _summary_report_indices(df, add_cols=None, add_index=None):
    if 'opset' not in df.columns:
        raise RuntimeError(  # pragma: no cover
            "Unable to create summary (opset missing)\n{}\n--\n{}".format(
                df.columns, df.head()))

    col_values = ["available"]
    for col in ['problem', 'scenario', 'opset', 'optim']:
        if col not in df.columns:
            df[col] = '' if col != 'opset' else numpy.nan
    indices = ["name", "problem", "scenario", 'optim', 'method_name',
               'output_index', 'conv_options', 'inst']
    indices = [i for i in indices if i in df.columns]
    df["optim"] = df["optim"].fillna('')
    for c in ['n_features', 'runtime']:
        if c in df.columns:
            indices.append(c)
            if c == 'runtime':
                df[c].fillna('-', inplace=True)
    for c in df.columns:
        if c.startswith('opset') or c in {'available'}:
            df[c].fillna('?', inplace=True)

    # Adds information about the models in the index
    indices2 = []
    for c in df.columns:
        if (isinstance(c, str) and len(c) >= 5 and (
                c.startswith("onx_") or c.startswith("skl_"))):
            if c in {'onx_domain', 'onx_doc_string', 'onx_ir_version',
                     'onx_model_version'}:
                continue
            if df[c].dtype in (numpy.float32, numpy.float64, float,
                               int, numpy.int32, numpy.int64):
                defval = -1
            else:
                defval = ''
            df[c].fillna(defval, inplace=True)
            if c.startswith('skl_'):
                indices.append(c)
            else:
                indices2.append(c)

    columns = ['opset']
    indices = indices + indices2
    if add_index is not None:
        for i in add_index:  # pragma: no cover
            if i not in indices:
                indices.append(i)
    return columns, indices, col_values


class _MyEncoder(json.JSONEncoder):
    def default(self, o):  # pylint: disable=E0202
        if hasattr(o, 'get_params'):
            obj = dict(clsname=o.__class__.__name__)
            obj.update(o.get_params())
            return json.dumps(obj, sort_keys=True)
        return json.dumps(o, sort_keys=True)  # pragma: no cover


def _jsonify(x):

    def _l(k):
        if isinstance(k, type):
            return k.__name__
        return k

    if isinstance(x, dict):
        x = {str(_l(k)): v for k, v in x.items()}
        try:
            return json.dumps(x, sort_keys=True, cls=_MyEncoder)
        except TypeError:  # pragma: no cover
            # Cannot sort.
            return json.dumps(x, cls=_MyEncoder)
    try:
        if numpy.isnan(x):
            x = ''
    except (ValueError, TypeError):
        pass
    try:
        return json.dumps(x, cls=_MyEncoder)
    except TypeError:  # pragma: no cover
        # Cannot sort.
        return json.dumps(x, cls=_MyEncoder)


def summary_report(df, add_cols=None, add_index=None):
    """
    Finalizes the results computed by function
    @see fn enumerate_validated_operator_opsets.

    @param      df          dataframe
    @param      add_cols    additional columns to take into account
                            as values
    @param      add_index   additional columns to take into accound
                            as index
    @return                 pivoted dataframe

    The outcome can be seen at page about :ref:`l-onnx-pyrun`.
    """
    df = df.copy()
    if 'inst' in df.columns:
        df['inst'] = df['inst'].apply(_jsonify)
    if 'conv_options' in df.columns:
        df['conv_options'] = df['conv_options'].apply(_jsonify)
    num_types = (int, float, decimal.Decimal, numpy.number)

    def aggfunc(values):
        if len(values) != 1:
            if all(map(lambda x: isinstance(x, num_types),
                       values)):
                mi, ma = min(values), max(values)
                if numpy.isnan(mi) and numpy.isnan(ma):
                    return ""
                if mi == ma:
                    return mi
                return '[{},{}]'.format(mi, ma)
            values = [str(_).replace("\n", " ").replace('\r', '').strip(" ")
                      for _ in values]
            values = [_ for _ in values if _]
            vals = set(values)
            if len(vals) != 1:
                return " // ".join(map(str, values))
        val = values.iloc[0] if not isinstance(values, list) else values[0]
        if isinstance(val, float) and numpy.isnan(val):
            return ""
        return str(val)

    columns, indices, col_values = _summary_report_indices(
        df, add_cols=add_cols, add_index=add_index)
    try:
        piv = pandas.pivot_table(df, values=col_values,
                                 index=indices, columns=columns,
                                 aggfunc=aggfunc).reset_index(drop=False)
    except (KeyError, TypeError) as e:  # pragma: no cover
        raise RuntimeError(
            "Issue with keys={}, values={}\namong {}.".format(
                indices, col_values, df.columns)) from e

    cols = list(piv.columns)
    opsets = [c[1] for c in cols if isinstance(c[1], (int, float))]

    versions = ["opset%d" % i for i in opsets]
    last = piv.columns[-1]
    if isinstance(last, tuple) and last == ('available', '?'):
        versions.append('FAIL')
    nbvalid = len(indices + versions)
    if len(piv.columns) != nbvalid:
        raise RuntimeError(  # pragma: no cover
            "Mismatch between {} != {}\n{}\n{}\n---\n{}\n{}\n{}".format(
                len(piv.columns), len(indices + versions),
                piv.columns, indices + versions,
                df.columns, indices, col_values))
    piv.columns = indices + versions
    piv = piv[indices + list(reversed(versions))].copy()
    for c in versions:
        piv[c].fillna('-', inplace=True)

    if "available-ERROR" in df.columns:

        from skl2onnx.common.exceptions import MissingShapeCalculator

        def replace_msg(text):
            if isinstance(text, MissingShapeCalculator):
                return "NO CONVERTER"  # pragma: no cover
            if str(text).startswith("Unable to find a shape calculator for type '"):
                return "NO CONVERTER"
            if str(text).startswith("Unable to find problem for model '"):
                return "NO PROBLEM"  # pragma: no cover
            if "not implemented for float64" in str(text):
                return "NO RUNTIME 64"  # pragma: no cover
            return str(text)

        piv2 = pandas.pivot_table(
            df, values="available-ERROR", index=indices,
            columns='opset', aggfunc=aggfunc).reset_index(drop=False)

        col = piv2.iloc[:, piv2.shape[1] - 1]
        piv["ERROR-msg"] = col.apply(replace_msg)

    if any('time-ratio-' in c for c in df.columns):
        cols = [c for c in df.columns if c.startswith('time-ratio')]
        cols.sort()

        df_sub = df[indices + cols]
        piv2 = df_sub.groupby(indices).mean()
        piv = piv.merge(piv2, on=indices, how='left')

        def rep(c):
            if 'N=1' in c and 'N=10' not in c:
                return c.replace("time-ratio-", "RT/SKL-")
            else:
                return c.replace("time-ratio-", "")
        cols = [rep(c) for c in piv.columns]
        piv.columns = cols

        # min, max
        mins = [c for c in piv.columns if c.endswith('-min')]
        maxs = [c for c in piv.columns if c.endswith('-max')]
        combined = []
        for mi, ma in zip(mins, maxs):
            combined.append(mi)
            combined.append(ma)
        first = [c for c in piv.columns if c not in combined]
        piv = piv[first + combined]

    def clean_values(value):
        if not isinstance(value, str):
            return value  # pragma: no cover
        if "ERROR->=1000000" in value:
            value = "big-diff"
        elif "ERROR" in value:
            value = value.replace("ERROR-_", "")
            value = value.replace("_exc", "")
            value = "ERR: " + value
        elif "OK-" in value:
            value = value.replace("OK-", "OK ")
        elif "e<" in value:
            value = value.replace("-", " ")
        return value

    for c in piv.columns:
        if "opset" in c:
            piv[c] = piv[c].apply(clean_values)
        if 'optim' in c:
            piv[c] = piv[c].apply(_clean_values_optim)

    # adding versions
    def keep_values(x):
        if isinstance(x, float) and numpy.isnan(x):
            return False  # pragma: no cover
        return True

    col_versions = [c for c in df.columns if c.startswith("v_")]
    if len(col_versions) > 0:
        for c in col_versions:
            vals = set(filter(keep_values, df[c]))
            if len(vals) != 1:
                raise RuntimeError(  # pragma: no cover
                    "Columns '{}' has multiple values {}.".format(c, vals))
            piv[c] = list(vals)[0]

    return piv


def merge_benchmark(dfs, column='runtime', baseline=None, suffix='-base'):
    """
    Merges several benchmarks run with command line
    :ref:`validate_runtime <l-cmd-validate_runtime>`.

    @param      dfs         dictionary *{'prefix': dataframe}*
    @param      column      every value from this column is prefixed
                            by the given key in *dfs*
    @param      baseline    add baseline
    @param      suffix      suffix to add when comparing to the baseline
    @return                 merged dataframe
    """
    def add_prefix(prefix, v):
        if isinstance(v, str):
            return prefix + v
        return v  # pragma: no cover

    conc = []
    for k, df in dfs.items():
        if column not in df.columns:
            raise ValueError(
                "Unable to find column '{}' in {} (key='{}')".format(
                    column, df.columns, k))
        df = df.copy()
        df[column] = df[column].apply(lambda x: add_prefix(k, x))
        if 'inst' in df.columns:
            df['inst'] = df['inst'].fillna('')
        else:
            df['inst'] = ''
        conc.append(df)
    merged = pandas.concat(conc).reset_index(drop=True)
    if baseline is not None:
        def get_key(index):
            k = []
            for v in index:
                try:
                    if numpy.isnan(v):
                        continue  # pragma: no cover
                except (ValueError, TypeError):
                    pass
                k.append(v)
            return tuple(k)

        columns, indices, _ = _summary_report_indices(merged)
        indices = list(_ for _ in (indices + columns) if _ != 'runtime')
        try:
            bdata = merged[merged.runtime == baseline].drop(
                'runtime', axis=1).set_index(indices, verify_integrity=True)
        except ValueError as e:
            bdata2 = merged[indices + ['runtime']].copy()
            bdata2['count'] = 1
            n_rows = bdata2['count'].sum()
            gr = bdata2.groupby(indices + ['runtime'], as_index=False).sum(
            ).sort_values('count', ascending=False)
            n_rows2 = gr['count'].sum()
            one = gr.head()[:1]
            rows = merged.merge(one, on=indices + ['runtime'])[:2]
            for c in ['init-types', 'bench-skl', 'bench-batch', 'init_types', 'cl']:
                if c in rows.columns:
                    rows = rows.drop(c, axis=1)
            srows = rows.T.to_string(min_rows=100)
            raise ValueError(
                "(n_rows={}, n_rows2={}) Unable to group by {}.\n{}\n-------\n{}".format(
                    n_rows, n_rows2, indices, gr.T, srows)) from e
        if bdata.shape[0] == 0:
            raise RuntimeError(  # pragma: no cover
                "No result for baseline '{}'.".format(baseline))
        ratios = [c for c in merged.columns if c.startswith('time-ratio-')]
        indexed = {}
        for index in bdata.index:
            row = bdata.loc[index, :]
            key = get_key(index)
            indexed[key] = row[ratios]

        for i in range(merged.shape[0]):
            key = get_key(tuple(merged.loc[i, indices]))
            if key not in indexed:
                continue  # pragma: no cover
            value = indexed[key]
            for r in ratios:
                if r.endswith('-min') or r.endswith('-max'):
                    continue
                value2 = merged.loc[i, r]
                new_r = value2 / value[r]
                new_col = r + suffix
                if new_col not in merged.columns:
                    merged[new_col] = numpy.nan
                merged.loc[i, new_col] = new_r

    return merged



"""
@file
@brief Helpers for benchmarks.
"""
from timeit import Timer
import numpy


def measure_time(stmt, *x, repeat=5, number=5, div_by_number=True, first_run=True):
    """
    Measures a statement and returns the results as a dictionary.

    :param stmt: string
    :param *x: inputs
    :param repeat: average over *repeat* experiment
    :param number: number of executions in one row
    :param div_by_number: divide by the number of executions
    :param first_run: if True, runs the function once before measuring
    :return: dictionary

    See `Timer.repeat
    <https://docs.python.org/3/library/timeit.html?timeit.Timer.repeat>`_
    for a better understanding of parameter *repeat* and *number*.
    The function returns a duration corresponding to
    *number* times the execution of the main statement.
    """
    try:
        stmt(*x)
    except RuntimeError as e:  # pragma: no cover
        raise RuntimeError("{}-{}".format(type(x), x.dtype)) from e

    def fct():
        stmt(*x)

    if first_run:
        fct()
    tim = Timer(fct)
    res = numpy.array(tim.repeat(repeat=repeat, number=number))
    total = numpy.sum(res)
    if div_by_number:
        res /= number
    mean = numpy.mean(res)
    dev = numpy.mean(res ** 2)
    dev = max(0, (dev - mean**2)) ** 0.5
    mes = dict(average=mean, deviation=dev, min_exec=numpy.min(res),
               max_exec=numpy.max(res), repeat=repeat, number=number,
               total=total)
    return mes



"""
@file
@brief Command line about validation of prediction runtime.
"""
import os
import pickle
from logging import getLogger
import warnings
from pandas import read_csv
from skl2onnx.common.data_types import FloatTensorType, DoubleTensorType
from ..onnx_conv import to_onnx
from ..onnxrt import OnnxInference
from ..onnx_tools.optim import onnx_optimisations
from ..onnxrt.validate.validate_difference import measure_relative_difference
from ..onnx_conv import guess_schema_from_data, guess_schema_from_model


def convert_validate(pkl, data=None, schema=None,
                     method="predict", name='Y',
                     target_opset=None,
                     outonnx="model.onnx",
                     runtime='python', metric="l1med",
                     use_double=None, noshape=False,
                     optim='onnx', rewrite_ops=True,
                     options=None, fLOG=print, verbose=1,
                     register=True):
    """
    Converts a model stored in *pkl* file and measure the differences
    between the model and the ONNX predictions.

    :param pkl: pickle file
    :param data: data file, loaded with pandas,
        converted to a single array, the data is used to guess
        the schema if *schema* not specified
    :param schema: initial type of the model
    :param method: method to call
    :param name: output name
    :param target_opset: target opset
    :param outonnx: produced ONNX model
    :param runtime: runtime to use to compute predictions,
        'python', 'python_compiled',
        'onnxruntime1' or 'onnxruntime2'
    :param metric: the metric 'l1med' is given by function
        :func:`measure_relative_difference
        <mlprodict.onnxrt.validate.validate_difference.measure_relative_difference>`
    :param noshape: run the conversion with no shape information
    :param use_double: use double for the runtime if possible,
        two possible options, ``"float64"`` or ``'switch'``,
        the first option produces an ONNX file with doubles,
        the second option loads an ONNX file (float or double)
        and replaces matrices in ONNX with the matrices coming from
        the model, this second way is just for testing purposes
    :param optim: applies optimisations on the first ONNX graph,
        use 'onnx' to reduce the number of node Identity and
        redundant subgraphs
    :param rewrite_ops: rewrites some converters from skl2onnx
    :param options: additional options for conversion,
        dictionary as a string
    :param verbose: verbose level
    :param register: registers additional converters implemented by this package
    :param fLOG: logging function
    :return: a dictionary with the results

    .. cmdref::
        :title: Converts and compares an ONNX file
        :cmd: -m mlprodict convert_validate --help
        :lid: l-cmd-convert_validate

        The command converts and validates a :epkg:`scikit-learn` model.
        An example to check the prediction of a logistic regression.

        ::

            import os
            import pickle
            import pandas
            from sklearn.datasets import load_iris
            from sklearn.model_selection import train_test_split
            from sklearn.linear_model import LogisticRegression
            from mlprodict.__main__ import main
            from mlprodict.cli import convert_validate

            iris = load_iris()
            X, y = iris.data, iris.target
            X_train, X_test, y_train, _ = train_test_split(X, y, random_state=11)
            clr = LogisticRegression()
            clr.fit(X_train, y_train)

            pandas.DataFrame(X_test).to_csv("data.csv", index=False)
            with open("model.pkl", "wb") as f:
                pickle.dump(clr, f)

        And the command line to check the predictions
        using a command line.

        ::

            convert_validate --pkl model.pkl --data data.csv
                             --method predict,predict_proba
                             --name output_label,output_probability
                             --verbose 1
    """
    if fLOG is None:
        verbose = 0  # pragma: no cover
    if use_double not in (None, 'float64', 'switch'):
        raise ValueError(  # pragma: no cover
            "use_double must be either None, 'float64' or 'switch'")
    if optim == '':
        optim = None  # pragma: no cover
    if target_opset == '':
        target_opset = None  # pragma: no cover
    if verbose == 0:
        logger = getLogger('skl2onnx')
        logger.disabled = True
    if not os.path.exists(pkl):
        raise FileNotFoundError(  # pragma: no cover
            "Unable to find model '{}'.".format(pkl))
    if os.path.exists(outonnx):
        warnings.warn("File '{}' will be overwritten.".format(outonnx))
    if verbose > 0:
        fLOG("[convert_validate] load model '{}'".format(pkl))
    with open(pkl, "rb") as f:
        model = pickle.load(f)

    if use_double == 'float64':
        tensor_type = DoubleTensorType
    else:
        tensor_type = FloatTensorType
    if options in (None, ''):
        options = None
    else:
        from ..onnxrt.validate.validate_scenarios import (
            interpret_options_from_string)
        options = interpret_options_from_string(options)
        if verbose > 0:
            fLOG("[convert_validate] options={}".format(repr(options)))

    if register:
        from ..onnx_conv import (
            register_converters, register_rewritten_operators)
        register_converters()
        register_rewritten_operators()

    # data and schema
    if data is None or not os.path.exists(data):
        if schema is None:
            schema = guess_schema_from_model(model, tensor_type)
        if verbose > 0:
            fLOG("[convert_validate] model schema={}".format(schema))
        df = None
    else:
        if verbose > 0:
            fLOG("[convert_validate] load data '{}'".format(data))
        df = read_csv(data)
        if verbose > 0:
            fLOG("[convert_validate] convert data into matrix")
        if schema is None:
            schema = guess_schema_from_data(df, tensor_type)
        if schema is None:
            schema = [  # pragma: no cover
                ('X', tensor_type([None, df.shape[1]]))]
        if len(schema) == 1:
            df = df.values  # pylint: disable=E1101
        if verbose > 0:
            fLOG("[convert_validate] data schema={}".format(schema))

    if noshape:
        if verbose > 0:
            fLOG(  # pragma: no cover
                "[convert_validate] convert the model with no shape information")
        schema = [(name, col.__class__([None, None])) for name, col in schema]
        onx = to_onnx(
            model, initial_types=schema, rewrite_ops=rewrite_ops,
            target_opset=target_opset, options=options)
    else:
        if verbose > 0:
            fLOG("[convert_validate] convert the model with shapes")
        onx = to_onnx(
            model, initial_types=schema, target_opset=target_opset,
            rewrite_ops=rewrite_ops, options=options)

    if optim is not None:
        if verbose > 0:
            fLOG("[convert_validate] run optimisations '{}'".format(optim))
        onx = onnx_optimisations(onx, optim=optim)
    if verbose > 0:
        fLOG("[convert_validate] saves to '{}'".format(outonnx))
    memory = onx.SerializeToString()
    with open(outonnx, 'wb') as f:
        f.write(memory)

    if verbose > 0:
        fLOG("[convert_validate] creates OnnxInference session")
    sess = OnnxInference(onx, runtime=runtime)
    if use_double == "switch":
        if verbose > 0:
            fLOG("[convert_validate] switch to double")
        sess.switch_initializers_dtype(model)

    if verbose > 0:
        fLOG("[convert_validate] compute prediction from model")

    if ',' in method:
        methods = method.split(',')
    else:
        methods = [method]
    if ',' in name:
        names = name.split(',')
    else:
        names = [name]

    if len(names) != len(methods):
        raise ValueError(
            "Number of methods and outputs do not match: {}, {}".format(
                names, methods))

    if metric != 'l1med':
        raise ValueError(  # pragma: no cover
            "Unknown metric '{}'".format(metric))

    if df is None:
        # no test on data
        return dict(onnx=memory)

    if verbose > 0:
        fLOG("[convert_validate] compute predictions from ONNX with name '{}'"
             "".format(name))

    ort_preds = sess.run(
        {'X': df}, verbose=max(verbose - 1, 0), fLOG=fLOG)

    metrics = []
    out_skl_preds = []
    out_ort_preds = []
    for method_, name_ in zip(methods, names):
        if verbose > 0:
            fLOG("[convert_validate] compute predictions with method '{}'".format(
                method_))
        meth = getattr(model, method_)
        skl_pred = meth(df)
        out_skl_preds.append(df)

        if name_ not in ort_preds:
            raise KeyError(
                "Unable to find output name '{}' in {}".format(
                    name_, list(sorted(ort_preds))))

        ort_pred = ort_preds[name_]
        out_ort_preds.append(ort_pred)
        diff = measure_relative_difference(skl_pred, ort_pred)
        if verbose > 0:
            fLOG("[convert_validate] {}={}".format(metric, diff))
        metrics.append(diff)

    return dict(skl_pred=out_skl_preds, ort_pred=out_ort_preds,
                metrics=metrics, onnx=memory)
